import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, List, Optional, Tuple, Type, cast

from lark import Tree

from datahub.configuration.source_common import PlatformDetail
from datahub.emitter import mce_builder as builder
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.source.powerbi.config import (
    Constant,
    DataBricksPlatformDetail,
    DataPlatformPair,
    PowerBiDashboardSourceConfig,
    PowerBiDashboardSourceReport,
    PowerBIPlatformDetail,
    SupportedDataPlatform,
)
from datahub.ingestion.source.powerbi.dataplatform_instance_resolver import (
    AbstractDataPlatformInstanceResolver,
)
from datahub.ingestion.source.powerbi.m_query import native_sql_parser, tree_function
from datahub.ingestion.source.powerbi.m_query.data_classes import (
    DataAccessFunctionDetail,
    DataPlatformTable,
    FunctionName,
    IdentifierAccessor,
    Lineage,
    ReferencedTable,
)
from datahub.ingestion.source.powerbi.rest_api_wrapper.data_classes import Table
from datahub.sql_parsing.sqlglot_lineage import SqlParsingResult

logger = logging.getLogger(__name__)


def get_next_item(items: List[str], item: str) -> Optional[str]:
    if item in items:
        try:
            index = items.index(item)
            return items[index + 1]
        except IndexError:
            logger.debug(f'item:"{item}", not found in item-list: {items}')
    return None


def urn_to_lowercase(value: str, flag: bool) -> str:
    if flag is True:
        return value.lower()

    return value


def make_urn(
    config: PowerBiDashboardSourceConfig,
    platform_instance_resolver: AbstractDataPlatformInstanceResolver,
    data_platform_pair: DataPlatformPair,
    server: str,
    qualified_table_name: str,
) -> str:
    platform_detail: PlatformDetail = platform_instance_resolver.get_platform_instance(
        PowerBIPlatformDetail(
            data_platform_pair=data_platform_pair,
            data_platform_server=server,
        )
    )

    return builder.make_dataset_urn_with_platform_instance(
        platform=data_platform_pair.datahub_data_platform_name,
        platform_instance=platform_detail.platform_instance,
        env=platform_detail.env,
        name=urn_to_lowercase(
            qualified_table_name, config.convert_lineage_urns_to_lowercase
        ),
    )


class AbstractLineage(ABC):
    """
    Base class to share common functionalities among different dataplatform for M-Query parsing.

    To create qualified table name we need to parse M-Query data-access-functions(https://learn.microsoft.com/en-us/powerquery-m/accessing-data-functions) and
    the data-access-functions has some define pattern to access database-name, schema-name and table-name, for example, see below M-Query.

        let
            Source = Sql.Database("localhost", "library"),
            dbo_book_issue = Source{[Schema="dbo",Item="book_issue"]}[Data]
        in
            dbo_book_issue

    It is MSSQL M-Query and Sql.Database is the data-access-function to access MSSQL. If this function is available in M-Query then database name is available in the second argument of the first statement and schema-name and table-name is available in the second statement. the second statement can be repeated to access different tables from MSSQL.

    DefaultTwoStepDataAccessSources extends the AbstractDataPlatformTableCreator and provides the common functionalities for data-platform which has above type of M-Query pattern

    data-access-function varies as per data-platform for example for MySQL.Database for MySQL, PostgreSQL.Database for Postgres and Oracle.Database for Oracle and number of statement to
    find out database-name , schema-name and table-name also varies as per dataplatform.

    Value.NativeQuery is one of the functions which is used to execute a native query inside M-Query, for example see below M-Query

        let
            Source = Value.NativeQuery(AmazonRedshift.Database("redshift-url","dev"), "select * from dev.public.category", null, [EnableFolding=true])
        in
            Source

    In this M-Query database-name is available in first argument and rest of the detail i.e database & schema is available in native query.

    NativeQueryDataPlatformTableCreator extends AbstractDataPlatformTableCreator to support Redshift and Snowflake native query parsing.

    """

    ctx: PipelineContext
    table: Table
    config: PowerBiDashboardSourceConfig
    reporter: PowerBiDashboardSourceReport
    platform_instance_resolver: AbstractDataPlatformInstanceResolver

    def __init__(
        self,
        ctx: PipelineContext,
        table: Table,
        config: PowerBiDashboardSourceConfig,
        reporter: PowerBiDashboardSourceReport,
        platform_instance_resolver: AbstractDataPlatformInstanceResolver,
    ) -> None:
        super().__init__()
        self.ctx = ctx
        self.table = table
        self.config = config
        self.reporter = reporter
        self.platform_instance_resolver = platform_instance_resolver

    @abstractmethod
    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        pass

    @abstractmethod
    def get_platform_pair(self) -> DataPlatformPair:
        pass

    @staticmethod
    def get_db_detail_from_argument(
        arg_list: Tree,
    ) -> Tuple[Optional[str], Optional[str]]:
        arguments: List[str] = tree_function.strip_char_from_list(
            values=tree_function.remove_whitespaces_from_list(
                tree_function.token_values(arg_list)
            ),
        )

        if len(arguments) < 2:
            logger.debug(f"Expected minimum 2 arguments, but got {len(arguments)}")
            return None, None

        return arguments[0], arguments[1]

    @staticmethod
    def create_reference_table(
        arg_list: Tree,
        table_detail: Dict[str, str],
    ) -> Optional[ReferencedTable]:
        arguments: List[str] = tree_function.strip_char_from_list(
            values=tree_function.remove_whitespaces_from_list(
                tree_function.token_values(arg_list)
            ),
        )

        logger.debug(f"Processing arguments {arguments}")

        if (
            len(arguments) >= 4  # [0] is warehouse FQDN.
            # [1] is endpoint, we are not using it.
            # [2] is "Catalog" key
            # [3] is catalog's value
        ):
            return ReferencedTable(
                warehouse=arguments[0],
                catalog=arguments[3],
                # As per my observation, database and catalog names are same in M-Query
                database=table_detail["Database"]
                if table_detail.get("Database")
                else arguments[3],
                schema=table_detail["Schema"],
                table=table_detail.get("Table") or table_detail["View"],
            )
        elif len(arguments) == 2:
            return ReferencedTable(
                warehouse=arguments[0],
                database=table_detail["Database"],
                schema=table_detail["Schema"],
                table=table_detail.get("Table") or table_detail["View"],
                catalog=None,
            )

        return None

    def parse_custom_sql(
        self, query: str, server: str, database: Optional[str], schema: Optional[str]
    ) -> Lineage:
        dataplatform_tables: List[DataPlatformTable] = []

        platform_detail: PlatformDetail = (
            self.platform_instance_resolver.get_platform_instance(
                PowerBIPlatformDetail(
                    data_platform_pair=self.get_platform_pair(),
                    data_platform_server=server,
                )
            )
        )

        query = native_sql_parser.remove_drop_statement(
            native_sql_parser.remove_special_characters(query)
        )

        parsed_result: Optional["SqlParsingResult"] = (
            native_sql_parser.parse_custom_sql(
                ctx=self.ctx,
                query=query,
                platform=self.get_platform_pair().datahub_data_platform_name,
                platform_instance=platform_detail.platform_instance,
                env=platform_detail.env,
                database=database,
                schema=schema,
            )
        )

        if parsed_result is None:
            self.reporter.info(
                title=Constant.SQL_PARSING_FAILURE,
                message="Fail to parse native sql present in PowerBI M-Query",
                context=f"table-name={self.table.full_name}, sql={query}",
            )
            return Lineage.empty()

        if parsed_result.debug_info and parsed_result.debug_info.table_error:
            self.reporter.warning(
                title=Constant.SQL_PARSING_FAILURE,
                message="Fail to parse native sql present in PowerBI M-Query",
                context=f"table-name={self.table.full_name}, error={parsed_result.debug_info.table_error},sql={query}",
            )
            return Lineage.empty()

        for urn in parsed_result.in_tables:
            dataplatform_tables.append(
                DataPlatformTable(
                    data_platform_pair=self.get_platform_pair(),
                    urn=urn,
                )
            )

        logger.debug(f"Native Query parsed result={parsed_result}")
        logger.debug(f"Generated dataplatform_tables={dataplatform_tables}")

        return Lineage(
            upstreams=dataplatform_tables,
            column_lineage=(
                parsed_result.column_lineage
                if parsed_result.column_lineage is not None
                else []
            ),
        )


class AmazonRedshiftLineage(AbstractLineage):
    def get_platform_pair(self) -> DataPlatformPair:
        return SupportedDataPlatform.AMAZON_REDSHIFT.value

    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        logger.debug(
            f"Processing AmazonRedshift data-access function detail {data_access_func_detail}"
        )

        server, db_name = self.get_db_detail_from_argument(
            data_access_func_detail.arg_list
        )
        if db_name is None or server is None:
            return Lineage.empty()  # Return an empty list

        schema_name: str = cast(
            IdentifierAccessor, data_access_func_detail.identifier_accessor
        ).items["Name"]

        table_name: str = cast(
            IdentifierAccessor,
            cast(IdentifierAccessor, data_access_func_detail.identifier_accessor).next,
        ).items["Name"]

        qualified_table_name: str = f"{db_name}.{schema_name}.{table_name}"

        urn = make_urn(
            config=self.config,
            platform_instance_resolver=self.platform_instance_resolver,
            data_platform_pair=self.get_platform_pair(),
            server=server,
            qualified_table_name=qualified_table_name,
        )

        return Lineage(
            upstreams=[
                DataPlatformTable(
                    data_platform_pair=self.get_platform_pair(),
                    urn=urn,
                )
            ],
            column_lineage=[],
        )


class OracleLineage(AbstractLineage):
    def get_platform_pair(self) -> DataPlatformPair:
        return SupportedDataPlatform.ORACLE.value

    @staticmethod
    def _get_server_and_db_name(value: str) -> Tuple[Optional[str], Optional[str]]:
        error_message: str = (
            f"The target argument ({value}) should in the format of <host-name>:<port>/<db-name>["
            ".<domain>]"
        )
        splitter_result: List[str] = value.split("/")
        if len(splitter_result) != 2:
            logger.debug(error_message)
            return None, None

        db_name = splitter_result[1].split(".")[0]

        return tree_function.strip_char_from_list([splitter_result[0]])[0], db_name

    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        logger.debug(
            f"Processing Oracle data-access function detail {data_access_func_detail}"
        )

        arguments: List[str] = tree_function.remove_whitespaces_from_list(
            tree_function.token_values(data_access_func_detail.arg_list)
        )

        server, db_name = self._get_server_and_db_name(arguments[0])

        if db_name is None or server is None:
            return Lineage.empty()

        schema_name: str = cast(
            IdentifierAccessor, data_access_func_detail.identifier_accessor
        ).items["Schema"]

        table_name: str = cast(
            IdentifierAccessor,
            cast(IdentifierAccessor, data_access_func_detail.identifier_accessor).next,
        ).items["Name"]

        qualified_table_name: str = f"{db_name}.{schema_name}.{table_name}"

        urn = make_urn(
            config=self.config,
            platform_instance_resolver=self.platform_instance_resolver,
            data_platform_pair=self.get_platform_pair(),
            server=server,
            qualified_table_name=qualified_table_name,
        )

        return Lineage(
            upstreams=[
                DataPlatformTable(
                    data_platform_pair=self.get_platform_pair(),
                    urn=urn,
                )
            ],
            column_lineage=[],
        )


class DatabricksLineage(AbstractLineage):
    def form_qualified_table_name(
        self,
        table_reference: ReferencedTable,
        data_platform_pair: DataPlatformPair,
    ) -> str:
        platform_detail: PlatformDetail = (
            self.platform_instance_resolver.get_platform_instance(
                PowerBIPlatformDetail(
                    data_platform_pair=data_platform_pair,
                    data_platform_server=table_reference.warehouse,
                )
            )
        )

        metastore: Optional[str] = None

        qualified_table_name: str = f"{table_reference.database}.{table_reference.schema}.{table_reference.table}"

        if isinstance(platform_detail, DataBricksPlatformDetail):
            metastore = platform_detail.metastore

        if metastore is not None:
            return f"{metastore}.{qualified_table_name}"

        return qualified_table_name

    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        logger.debug(
            f"Processing Databrick data-access function detail {data_access_func_detail}"
        )
        table_detail: Dict[str, str] = {}
        temp_accessor: Optional[IdentifierAccessor] = (
            data_access_func_detail.identifier_accessor
        )

        while temp_accessor:
            # Condition to handle databricks M-query pattern where table, schema and database all are present in
            # the same invoke statement
            if all(
                element in temp_accessor.items
                for element in ["Item", "Schema", "Catalog"]
            ):
                table_detail["Schema"] = temp_accessor.items["Schema"]
                table_detail["Table"] = temp_accessor.items["Item"]
            else:
                table_detail[temp_accessor.items["Kind"]] = temp_accessor.items["Name"]

            if temp_accessor.next is not None:
                temp_accessor = temp_accessor.next
            else:
                break

        table_reference = self.create_reference_table(
            arg_list=data_access_func_detail.arg_list,
            table_detail=table_detail,
        )

        if table_reference:
            qualified_table_name: str = self.form_qualified_table_name(
                table_reference=table_reference,
                data_platform_pair=self.get_platform_pair(),
            )

            urn = make_urn(
                config=self.config,
                platform_instance_resolver=self.platform_instance_resolver,
                data_platform_pair=self.get_platform_pair(),
                server=table_reference.warehouse,
                qualified_table_name=qualified_table_name,
            )

            return Lineage(
                upstreams=[
                    DataPlatformTable(
                        data_platform_pair=self.get_platform_pair(),
                        urn=urn,
                    )
                ],
                column_lineage=[],
            )

        return Lineage.empty()

    def get_platform_pair(self) -> DataPlatformPair:
        return SupportedDataPlatform.DATABRICKS_SQL.value


class TwoStepDataAccessPattern(AbstractLineage, ABC):
    """
    These are the DataSource for which PowerBI Desktop generates default M-Query of the following pattern
        let
            Source = Sql.Database("localhost", "library"),
            dbo_book_issue = Source{[Schema="dbo",Item="book_issue"]}[Data]
        in
            dbo_book_issue
    """

    def two_level_access_pattern(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        logger.debug(
            f"Processing {self.get_platform_pair().powerbi_data_platform_name} data-access function detail {data_access_func_detail}"
        )

        server, db_name = self.get_db_detail_from_argument(
            data_access_func_detail.arg_list
        )
        if server is None or db_name is None:
            return Lineage.empty()  # Return an empty list

        schema_name: str = cast(
            IdentifierAccessor, data_access_func_detail.identifier_accessor
        ).items["Schema"]

        table_name: str = cast(
            IdentifierAccessor, data_access_func_detail.identifier_accessor
        ).items["Item"]

        qualified_table_name: str = f"{db_name}.{schema_name}.{table_name}"

        logger.debug(
            f"Platform({self.get_platform_pair().datahub_data_platform_name}) qualified_table_name= {qualified_table_name}"
        )

        urn = make_urn(
            config=self.config,
            platform_instance_resolver=self.platform_instance_resolver,
            data_platform_pair=self.get_platform_pair(),
            server=server,
            qualified_table_name=qualified_table_name,
        )
        return Lineage(
            upstreams=[
                DataPlatformTable(
                    data_platform_pair=self.get_platform_pair(),
                    urn=urn,
                )
            ],
            column_lineage=[],
        )


class PostgresLineage(TwoStepDataAccessPattern):
    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        return self.two_level_access_pattern(data_access_func_detail)

    def get_platform_pair(self) -> DataPlatformPair:
        return SupportedDataPlatform.POSTGRES_SQL.value


class MSSqlLineage(TwoStepDataAccessPattern):
    # https://learn.microsoft.com/en-us/sql/relational-databases/security/authentication-access/ownership-and-user-schema-separation?view=sql-server-ver16
    DEFAULT_SCHEMA = "dbo"  # Default schema name in MS-SQL is dbo

    def get_platform_pair(self) -> DataPlatformPair:
        return SupportedDataPlatform.MS_SQL.value

    def create_urn_using_old_parser(
        self, query: str, db_name: str, server: str
    ) -> List[DataPlatformTable]:
        dataplatform_tables: List[DataPlatformTable] = []

        tables: List[str] = native_sql_parser.get_tables(query)

        for parsed_table in tables:
            # components: List[str] = [v.strip("[]") for v in parsed_table.split(".")]
            components = [v.strip("[]") for v in parsed_table.split(".")]
            if len(components) == 3:
                database, schema, table = components
            elif len(components) == 2:
                schema, table = components
                database = db_name
            elif len(components) == 1:
                (table,) = components
                database = db_name
                schema = MSSqlLineage.DEFAULT_SCHEMA
            else:
                self.reporter.warning(
                    title="Invalid table format",
                    message="The advanced SQL lineage feature (enable_advance_lineage_sql_construct) is disabled. Please either enable this feature or ensure the table is referenced as <db-name>.<schema-name>.<table-name> in the SQL.",
                    context=f"table-name={self.table.full_name}",
                )
                continue

            qualified_table_name = f"{database}.{schema}.{table}"
            urn = make_urn(
                config=self.config,
                platform_instance_resolver=self.platform_instance_resolver,
                data_platform_pair=self.get_platform_pair(),
                server=server,
                qualified_table_name=qualified_table_name,
            )
            dataplatform_tables.append(
                DataPlatformTable(
                    data_platform_pair=self.get_platform_pair(),
                    urn=urn,
                )
            )

        logger.debug(f"Generated upstream tables = {dataplatform_tables}")

        return dataplatform_tables

    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        arguments: List[str] = tree_function.strip_char_from_list(
            values=tree_function.remove_whitespaces_from_list(
                tree_function.token_values(data_access_func_detail.arg_list)
            ),
        )

        server, database = self.get_db_detail_from_argument(
            data_access_func_detail.arg_list
        )
        if server is None or database is None:
            return Lineage.empty()  # Return an empty list

        assert server
        assert database  # to silent the lint

        query: Optional[str] = get_next_item(arguments, "Query")
        if query:
            if self.config.enable_advance_lineage_sql_construct is False:
                # Use previous parser to generate URN to keep backward compatibility
                return Lineage(
                    upstreams=self.create_urn_using_old_parser(
                        query=query,
                        db_name=database,
                        server=server,
                    ),
                    column_lineage=[],
                )

            return self.parse_custom_sql(
                query=query,
                database=database,
                server=server,
                schema=MSSqlLineage.DEFAULT_SCHEMA,
            )

        # It is a regular case of MS-SQL
        logger.debug("Handling with regular case")
        return self.two_level_access_pattern(data_access_func_detail)


class ThreeStepDataAccessPattern(AbstractLineage, ABC):
    def get_datasource_server(
        self, arguments: List[str], data_access_func_detail: DataAccessFunctionDetail
    ) -> str:
        return tree_function.strip_char_from_list([arguments[0]])[0]

    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        logger.debug(
            f"Processing {self.get_platform_pair().datahub_data_platform_name} function detail {data_access_func_detail}"
        )

        arguments: List[str] = tree_function.remove_whitespaces_from_list(
            tree_function.token_values(data_access_func_detail.arg_list)
        )
        # First is database name
        db_name: str = data_access_func_detail.identifier_accessor.items["Name"]  # type: ignore
        # Second is schema name
        schema_name: str = cast(
            IdentifierAccessor,
            data_access_func_detail.identifier_accessor.next,  # type: ignore
        ).items["Name"]
        # Third is table name
        table_name: str = cast(
            IdentifierAccessor,
            data_access_func_detail.identifier_accessor.next.next,  # type: ignore
        ).items["Name"]

        qualified_table_name: str = f"{db_name}.{schema_name}.{table_name}"

        logger.debug(
            f"{self.get_platform_pair().datahub_data_platform_name} qualified_table_name {qualified_table_name}"
        )

        server: str = self.get_datasource_server(arguments, data_access_func_detail)

        urn = make_urn(
            config=self.config,
            platform_instance_resolver=self.platform_instance_resolver,
            data_platform_pair=self.get_platform_pair(),
            server=server,
            qualified_table_name=qualified_table_name,
        )

        return Lineage(
            upstreams=[
                DataPlatformTable(
                    data_platform_pair=self.get_platform_pair(),
                    urn=urn,
                )
            ],
            column_lineage=[],
        )


class SnowflakeLineage(ThreeStepDataAccessPattern):
    def get_platform_pair(self) -> DataPlatformPair:
        return SupportedDataPlatform.SNOWFLAKE.value


class GoogleBigQueryLineage(ThreeStepDataAccessPattern):
    def get_platform_pair(self) -> DataPlatformPair:
        return SupportedDataPlatform.GOOGLE_BIGQUERY.value

    def get_datasource_server(
        self, arguments: List[str], data_access_func_detail: DataAccessFunctionDetail
    ) -> str:
        # In Google BigQuery server is project-name
        # condition to silent lint, it is not going to be None
        return (
            data_access_func_detail.identifier_accessor.items["Name"]
            if data_access_func_detail.identifier_accessor is not None
            else ""
        )


class NativeQueryLineage(AbstractLineage):
    SUPPORTED_NATIVE_QUERY_DATA_PLATFORM: dict = {
        SupportedDataPlatform.SNOWFLAKE.value.powerbi_data_platform_name: SupportedDataPlatform.SNOWFLAKE,
        SupportedDataPlatform.AMAZON_REDSHIFT.value.powerbi_data_platform_name: SupportedDataPlatform.AMAZON_REDSHIFT,
        SupportedDataPlatform.DatabricksMultiCloud_SQL.value.powerbi_data_platform_name: SupportedDataPlatform.DatabricksMultiCloud_SQL,
    }
    current_data_platform: SupportedDataPlatform = SupportedDataPlatform.SNOWFLAKE

    def get_platform_pair(self) -> DataPlatformPair:
        return self.current_data_platform.value

    @staticmethod
    def is_native_parsing_supported(data_access_function_name: str) -> bool:
        return (
            data_access_function_name
            in NativeQueryLineage.SUPPORTED_NATIVE_QUERY_DATA_PLATFORM
        )

    def create_urn_using_old_parser(self, query: str, server: str) -> Lineage:
        dataplatform_tables: List[DataPlatformTable] = []

        tables: List[str] = native_sql_parser.get_tables(query)

        for qualified_table_name in tables:
            if len(qualified_table_name.split(".")) != 3:
                logger.debug(
                    f"Skipping table {qualified_table_name} as it is not as per qualified_table_name format"
                )
                continue

            urn = make_urn(
                config=self.config,
                platform_instance_resolver=self.platform_instance_resolver,
                data_platform_pair=self.get_platform_pair(),
                server=server,
                qualified_table_name=qualified_table_name,
            )

            dataplatform_tables.append(
                DataPlatformTable(
                    data_platform_pair=self.get_platform_pair(),
                    urn=urn,
                )
            )

        logger.debug(f"Generated dataplatform_tables {dataplatform_tables}")

        return Lineage(
            upstreams=dataplatform_tables,
            column_lineage=[],
        )

    def get_db_name(self, data_access_tokens: List[str]) -> Optional[str]:
        if (
            data_access_tokens[0]
            != SupportedDataPlatform.DatabricksMultiCloud_SQL.value.powerbi_data_platform_name
        ):
            return None

        database: Optional[str] = get_next_item(data_access_tokens, "Database")

        if (
            database and database != Constant.M_QUERY_NULL
        ):  # database name is explicitly set
            return database

        return (
            get_next_item(  # database name is set in Name argument
                data_access_tokens, "Name"
            )
            or get_next_item(  # If both above arguments are not available, then try Catalog
                data_access_tokens, "Catalog"
            )
        )

    def create_lineage(
        self, data_access_func_detail: DataAccessFunctionDetail
    ) -> Lineage:
        t1: Optional[Tree] = tree_function.first_arg_list_func(
            data_access_func_detail.arg_list
        )
        assert t1 is not None
        flat_argument_list: List[Tree] = tree_function.flat_argument_list(t1)

        if len(flat_argument_list) != 2:
            logger.debug(
                f"Expecting 2 argument, actual argument count is {len(flat_argument_list)}"
            )
            logger.debug(f"Flat argument list = {flat_argument_list}")
            return Lineage.empty()

        data_access_tokens: List[str] = tree_function.remove_whitespaces_from_list(
            tree_function.token_values(flat_argument_list[0])
        )

        if not self.is_native_parsing_supported(data_access_tokens[0]):
            logger.debug(
                f"Unsupported native-query data-platform = {data_access_tokens[0]}"
            )
            logger.debug(
                f"NativeQuery is supported only for {self.SUPPORTED_NATIVE_QUERY_DATA_PLATFORM}"
            )

            return Lineage.empty()

        if len(data_access_tokens[0]) < 3:
            logger.debug(
                f"Server is not available in argument list for data-platform {data_access_tokens[0]}. Returning empty "
                "list"
            )
            return Lineage.empty()

        self.current_data_platform = self.SUPPORTED_NATIVE_QUERY_DATA_PLATFORM[
            data_access_tokens[0]
        ]
        # The First argument is the query
        sql_query: str = tree_function.strip_char_from_list(
            values=tree_function.remove_whitespaces_from_list(
                tree_function.token_values(flat_argument_list[1])
            ),
        )[0]  # Remove any whitespaces and double quotes character

        server = tree_function.strip_char_from_list([data_access_tokens[2]])[0]

        if self.config.enable_advance_lineage_sql_construct is False:
            # Use previous parser to generate URN to keep backward compatibility
            return self.create_urn_using_old_parser(
                query=sql_query,
                server=server,
            )

        database_name: Optional[str] = self.get_db_name(data_access_tokens)

        return self.parse_custom_sql(
            query=sql_query,
            server=server,
            database=database_name,
            schema=None,
        )


class SupportedPattern(Enum):
    DATABRICKS_QUERY = (
        DatabricksLineage,
        FunctionName.DATABRICK_DATA_ACCESS,
    )

    DATABRICKS_MULTI_CLOUD = (
        DatabricksLineage,
        FunctionName.DATABRICK_MULTI_CLOUD_DATA_ACCESS,
    )

    POSTGRES_SQL = (
        PostgresLineage,
        FunctionName.POSTGRESQL_DATA_ACCESS,
    )

    ORACLE = (
        OracleLineage,
        FunctionName.ORACLE_DATA_ACCESS,
    )

    SNOWFLAKE = (
        SnowflakeLineage,
        FunctionName.SNOWFLAKE_DATA_ACCESS,
    )

    MS_SQL = (
        MSSqlLineage,
        FunctionName.MSSQL_DATA_ACCESS,
    )

    GOOGLE_BIG_QUERY = (
        GoogleBigQueryLineage,
        FunctionName.GOOGLE_BIGQUERY_DATA_ACCESS,
    )

    AMAZON_REDSHIFT = (
        AmazonRedshiftLineage,
        FunctionName.AMAZON_REDSHIFT_DATA_ACCESS,
    )

    NATIVE_QUERY = (
        NativeQueryLineage,
        FunctionName.NATIVE_QUERY,
    )

    def handler(self) -> Type[AbstractLineage]:
        return self.value[0]

    def function_name(self) -> str:
        return self.value[1].value

    @staticmethod
    def get_function_names() -> List[str]:
        functions: List[str] = []
        for supported_resolver in SupportedPattern:
            functions.append(supported_resolver.function_name())

        return functions

    @staticmethod
    def get_pattern_handler(function_name: str) -> Optional["SupportedPattern"]:
        logger.debug(f"Looking for pattern-handler for {function_name}")
        for supported_resolver in SupportedPattern:
            if function_name == supported_resolver.function_name():
                return supported_resolver
        logger.debug(f"pattern-handler not found for function_name {function_name}")
        return None



"""Convenience functions for creating MCEs"""

import hashlib
import json
import logging
import os
import re
import time
from datetime import datetime, timezone
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
    overload,
)

import typing_inspect
from avrogen.dict_wrapper import DictWrapper
from typing_extensions import assert_never

from datahub.emitter.enum_helpers import get_enum_options
from datahub.metadata.schema_classes import (
    AssertionKeyClass,
    AuditStampClass,
    ChartKeyClass,
    ContainerKeyClass,
    DashboardKeyClass,
    DatasetKeyClass,
    DatasetLineageTypeClass,
    DatasetSnapshotClass,
    FabricTypeClass,
    GlobalTagsClass,
    GlossaryTermAssociationClass,
    GlossaryTermsClass as GlossaryTerms,
    MetadataChangeEventClass,
    OwnerClass,
    OwnershipClass,
    OwnershipSourceClass,
    OwnershipSourceTypeClass,
    OwnershipTypeClass,
    SchemaFieldKeyClass,
    TagAssociationClass,
    UpstreamClass,
    UpstreamLineageClass,
    _Aspect as AspectAbstract,
)
from datahub.metadata.urns import DataFlowUrn, DatasetUrn, TagUrn
from datahub.utilities.urn_encoder import UrnEncoder

logger = logging.getLogger(__name__)
Aspect = TypeVar("Aspect", bound=AspectAbstract)

DEFAULT_ENV = FabricTypeClass.PROD
ALL_ENV_TYPES: Set[str] = set(get_enum_options(FabricTypeClass))

DEFAULT_FLOW_CLUSTER = "prod"
UNKNOWN_USER = "urn:li:corpuser:unknown"
DATASET_URN_TO_LOWER: bool = (
    os.getenv("DATAHUB_DATASET_URN_TO_LOWER", "false") == "true"
)

if TYPE_CHECKING:
    from datahub.emitter.mcp_builder import DatahubKey


# TODO: Delete this once lower-casing is the standard.
def set_dataset_urn_to_lower(value: bool) -> None:
    global DATASET_URN_TO_LOWER
    DATASET_URN_TO_LOWER = value


class OwnerType(Enum):
    USER = "corpuser"
    GROUP = "corpGroup"


def get_sys_time() -> int:
    # TODO deprecate this
    return int(time.time() * 1000)


@overload
def make_ts_millis(ts: None) -> None: ...


@overload
def make_ts_millis(ts: datetime) -> int: ...


def make_ts_millis(ts: Optional[datetime]) -> Optional[int]:
    # TODO: This duplicates the functionality of datetime_to_ts_millis
    if ts is None:
        return None
    return int(ts.timestamp() * 1000)


@overload
def parse_ts_millis(ts: float) -> datetime: ...


@overload
def parse_ts_millis(ts: None) -> None: ...


def parse_ts_millis(ts: Optional[float]) -> Optional[datetime]:
    if ts is None:
        return None
    return datetime.fromtimestamp(ts / 1000, tz=timezone.utc)


def make_data_platform_urn(platform: str) -> str:
    if platform.startswith("urn:li:dataPlatform:"):
        return platform
    return f"urn:li:dataPlatform:{platform}"


def make_dataset_urn(platform: str, name: str, env: str = DEFAULT_ENV) -> str:
    return make_dataset_urn_with_platform_instance(
        platform=platform, name=name, platform_instance=None, env=env
    )


def make_dataplatform_instance_urn(platform: str, instance: str) -> str:
    if instance.startswith("urn:li:dataPlatformInstance"):
        return instance
    else:
        return f"urn:li:dataPlatformInstance:({make_data_platform_urn(platform)},{instance})"


def make_dataset_urn_with_platform_instance(
    platform: str, name: str, platform_instance: Optional[str], env: str = DEFAULT_ENV
) -> str:
    if DATASET_URN_TO_LOWER:
        name = name.lower()
    return str(
        DatasetUrn.create_from_ids(
            platform_id=platform,
            table_name=name,
            env=env,
            platform_instance=platform_instance,
        )
    )


# Schema Field Urns url-encode reserved characters.
# TODO: This needs to be handled on consumer (UI) side well.
def make_schema_field_urn(parent_urn: str, field_path: str) -> str:
    assert parent_urn.startswith("urn:li:"), "Schema field's parent must be an urn"
    return f"urn:li:schemaField:({parent_urn},{UrnEncoder.encode_string(field_path)})"


def schema_field_urn_to_key(schema_field_urn: str) -> Optional[SchemaFieldKeyClass]:
    pattern = r"urn:li:schemaField:\((.*),(.*)\)"
    results = re.search(pattern, schema_field_urn)
    if results is not None:
        dataset_urn: str = results[1]
        field_path: str = results[2]
        return SchemaFieldKeyClass(parent=dataset_urn, fieldPath=field_path)
    return None


def dataset_urn_to_key(dataset_urn: str) -> Optional[DatasetKeyClass]:
    pattern = r"urn:li:dataset:\((.*),(.*),(.*)\)"
    results = re.search(pattern, dataset_urn)
    if results is not None:
        return DatasetKeyClass(platform=results[1], name=results[2], origin=results[3])
    return None


def dataset_key_to_urn(key: DatasetKeyClass) -> str:
    return f"urn:li:dataset:({key.platform},{key.name},{key.origin})"


def make_container_urn(guid: Union[str, "DatahubKey"]) -> str:
    if isinstance(guid, str) and guid.startswith("urn:li:container"):
        return guid
    else:
        from datahub.emitter.mcp_builder import DatahubKey

        if isinstance(guid, DatahubKey):
            guid = guid.guid()

        return f"urn:li:container:{guid}"


def container_urn_to_key(guid: str) -> Optional[ContainerKeyClass]:
    pattern = r"urn:li:container:(.*)"
    results = re.search(pattern, guid)
    if results is not None:
        return ContainerKeyClass(guid=results[1])
    return None


class _DatahubKeyJSONEncoder(json.JSONEncoder):
    # overload method default
    def default(self, obj: Any) -> Any:
        if hasattr(obj, "guid"):
            return obj.guid()
        # Call the default method for other types
        return json.JSONEncoder.default(self, obj)


def datahub_guid(obj: dict) -> str:
    json_key = json.dumps(
        obj,
        separators=(",", ":"),
        sort_keys=True,
        cls=_DatahubKeyJSONEncoder,
    )
    md5_hash = hashlib.md5(json_key.encode("utf-8"))
    return str(md5_hash.hexdigest())


def make_assertion_urn(assertion_id: str) -> str:
    return f"urn:li:assertion:{assertion_id}"


def assertion_urn_to_key(assertion_urn: str) -> Optional[AssertionKeyClass]:
    pattern = r"urn:li:assertion:(.*)"
    results = re.search(pattern, assertion_urn)
    if results is not None:
        return AssertionKeyClass(assertionId=results[1])
    return None


def make_user_urn(username: str) -> str:
    """
    Makes a user urn if the input is not a user or group urn already
    """
    return (
        f"urn:li:corpuser:{username}"
        if not username.startswith(("urn:li:corpuser:", "urn:li:corpGroup:"))
        else username
    )


def make_group_urn(groupname: str) -> str:
    """
    Makes a group urn if the input is not a user or group urn already
    """
    if groupname and groupname.startswith(("urn:li:corpGroup:", "urn:li:corpuser:")):
        return groupname
    else:
        return f"urn:li:corpGroup:{groupname}"


def make_tag_urn(tag: str) -> str:
    """
    Makes a tag urn if the input is not a tag urn already
    """
    if tag and tag.startswith("urn:li:tag:"):
        return tag
    return str(TagUrn(tag))


def make_owner_urn(owner: str, owner_type: OwnerType) -> str:
    if owner_type == OwnerType.USER:
        return make_user_urn(owner)
    elif owner_type == OwnerType.GROUP:
        return make_group_urn(owner)
    else:
        assert_never(owner_type)


def make_ownership_type_urn(type: str) -> str:
    return f"urn:li:ownershipType:{type}"


def make_term_urn(term: str) -> str:
    """
    Makes a term urn if the input is not a term urn already
    """
    if term and term.startswith("urn:li:glossaryTerm:"):
        return term
    else:
        return f"urn:li:glossaryTerm:{term}"


def make_data_flow_urn(
    orchestrator: str,
    flow_id: str,
    cluster: str = DEFAULT_FLOW_CLUSTER,
    platform_instance: Optional[str] = None,
) -> str:
    return str(
        DataFlowUrn.create_from_ids(
            orchestrator=orchestrator,
            flow_id=flow_id,
            env=cluster,
            platform_instance=platform_instance,
        )
    )


def make_data_job_urn_with_flow(flow_urn: str, job_id: str) -> str:
    return f"urn:li:dataJob:({flow_urn},{job_id})"


def make_data_process_instance_urn(dataProcessInstanceId: str) -> str:
    return f"urn:li:dataProcessInstance:{dataProcessInstanceId}"


def make_data_job_urn(
    orchestrator: str,
    flow_id: str,
    job_id: str,
    cluster: str = DEFAULT_FLOW_CLUSTER,
    platform_instance: Optional[str] = None,
) -> str:
    return make_data_job_urn_with_flow(
        make_data_flow_urn(orchestrator, flow_id, cluster, platform_instance), job_id
    )


def make_dashboard_urn(
    platform: str, name: str, platform_instance: Optional[str] = None
) -> str:
    # FIXME: dashboards don't currently include data platform urn prefixes.
    if platform_instance:
        return f"urn:li:dashboard:({platform},{platform_instance}.{name})"
    else:
        return f"urn:li:dashboard:({platform},{name})"


def dashboard_urn_to_key(dashboard_urn: str) -> Optional[DashboardKeyClass]:
    pattern = r"urn:li:dashboard:\((.*),(.*)\)"
    results = re.search(pattern, dashboard_urn)
    if results is not None:
        return DashboardKeyClass(dashboardTool=results[1], dashboardId=results[2])
    return None


def make_chart_urn(
    platform: str, name: str, platform_instance: Optional[str] = None
) -> str:
    # FIXME: charts don't currently include data platform urn prefixes.
    if platform_instance:
        return f"urn:li:chart:({platform},{platform_instance}.{name})"
    else:
        return f"urn:li:chart:({platform},{name})"


def chart_urn_to_key(chart_urn: str) -> Optional[ChartKeyClass]:
    pattern = r"urn:li:chart:\((.*),(.*)\)"
    results = re.search(pattern, chart_urn)
    if results is not None:
        return ChartKeyClass(dashboardTool=results[1], chartId=results[2])
    return None


def make_domain_urn(domain: str) -> str:
    if domain.startswith("urn:li:domain:"):
        return domain
    return f"urn:li:domain:{domain}"


def make_ml_primary_key_urn(feature_table_name: str, primary_key_name: str) -> str:
    return f"urn:li:mlPrimaryKey:({feature_table_name},{primary_key_name})"


def make_ml_feature_urn(
    feature_table_name: str,
    feature_name: str,
) -> str:
    return f"urn:li:mlFeature:({feature_table_name},{feature_name})"


def make_ml_feature_table_urn(platform: str, feature_table_name: str) -> str:
    return f"urn:li:mlFeatureTable:({make_data_platform_urn(platform)},{feature_table_name})"


def make_ml_model_urn(platform: str, model_name: str, env: str) -> str:
    return f"urn:li:mlModel:({make_data_platform_urn(platform)},{model_name},{env})"


def make_ml_model_deployment_urn(platform: str, deployment_name: str, env: str) -> str:
    return f"urn:li:mlModelDeployment:({make_data_platform_urn(platform)},{deployment_name},{env})"


def make_ml_model_group_urn(platform: str, group_name: str, env: str) -> str:
    return (
        f"urn:li:mlModelGroup:({make_data_platform_urn(platform)},{group_name},{env})"
    )


def validate_ownership_type(ownership_type: str) -> Tuple[str, Optional[str]]:
    if ownership_type.startswith("urn:li:"):
        return OwnershipTypeClass.CUSTOM, ownership_type
    ownership_type = ownership_type.upper()
    if ownership_type in get_enum_options(OwnershipTypeClass):
        return ownership_type, None
    raise ValueError(f"Unexpected ownership type: {ownership_type}")


def make_lineage_mce(
    upstream_urns: List[str],
    downstream_urn: str,
    lineage_type: str = DatasetLineageTypeClass.TRANSFORMED,
) -> MetadataChangeEventClass:
    """
    Note: this function only supports lineage for dataset aspects. It will not
    update lineage for any other aspect types.
    """
    mce = MetadataChangeEventClass(
        proposedSnapshot=DatasetSnapshotClass(
            urn=downstream_urn,
            aspects=[
                UpstreamLineageClass(
                    upstreams=[
                        UpstreamClass(
                            dataset=upstream_urn,
                            type=lineage_type,
                        )
                        for upstream_urn in upstream_urns
                    ]
                )
            ],
        )
    )
    return mce


def can_add_aspect_to_snapshot(
    SnapshotType: Type[DictWrapper], AspectType: Type[Aspect]
) -> bool:
    constructor_annotations = get_type_hints(SnapshotType.__init__)
    aspect_list_union = typing_inspect.get_args(constructor_annotations["aspects"])[0]

    supported_aspect_types = typing_inspect.get_args(aspect_list_union)

    return issubclass(AspectType, supported_aspect_types)


def can_add_aspect(mce: MetadataChangeEventClass, AspectType: Type[Aspect]) -> bool:
    # TODO: This is specific to snapshot types. We have a more general method
    # in `entity_supports_aspect`, which should be used instead. This method
    # should be deprecated, and all usages should be replaced.

    SnapshotType = type(mce.proposedSnapshot)

    return can_add_aspect_to_snapshot(SnapshotType, AspectType)


def assert_can_add_aspect(
    mce: MetadataChangeEventClass, AspectType: Type[Aspect]
) -> None:
    if not can_add_aspect(mce, AspectType):
        raise AssertionError(
            f"Cannot add aspect {AspectType} to {type(mce.proposedSnapshot)}"
        )


def get_aspect_if_available(
    mce: MetadataChangeEventClass, AspectType: Type[Aspect]
) -> Optional[Aspect]:
    assert_can_add_aspect(mce, AspectType)

    all_aspects = mce.proposedSnapshot.aspects
    aspects: List[Aspect] = [
        aspect for aspect in all_aspects if isinstance(aspect, AspectType)
    ]

    if len(aspects) > 1:
        raise ValueError(
            f"MCE contains multiple aspects of type {AspectType}: {aspects}"
        )
    if aspects:
        return aspects[0]
    return None


def remove_aspect_if_available(
    mce: MetadataChangeEventClass, aspect_type: Type[Aspect]
) -> bool:
    assert_can_add_aspect(mce, aspect_type)
    # loose type annotations since we checked before
    aspects: List[Any] = [
        aspect
        for aspect in mce.proposedSnapshot.aspects
        if not isinstance(aspect, aspect_type)
    ]
    removed = len(aspects) != len(mce.proposedSnapshot.aspects)
    mce.proposedSnapshot.aspects = aspects
    return removed


def get_or_add_aspect(mce: MetadataChangeEventClass, default: Aspect) -> Aspect:
    existing = get_aspect_if_available(mce, type(default))
    if existing is not None:
        return existing
    mce.proposedSnapshot.aspects.append(default)  # type: ignore
    return default


def make_global_tag_aspect_with_tag_list(tags: List[str]) -> GlobalTagsClass:
    return GlobalTagsClass(
        tags=[TagAssociationClass(make_tag_urn(tag)) for tag in tags]
    )


def make_ownership_aspect_from_urn_list(
    owner_urns: List[str],
    source_type: Optional[Union[str, OwnershipSourceTypeClass]],
    owner_type: Union[str, OwnershipTypeClass] = OwnershipTypeClass.DATAOWNER,
) -> OwnershipClass:
    for owner_urn in owner_urns:
        assert owner_urn.startswith("urn:li:corpuser:") or owner_urn.startswith(
            "urn:li:corpGroup:"
        )
    ownership_source_type: Union[None, OwnershipSourceClass] = None
    if source_type:
        ownership_source_type = OwnershipSourceClass(type=source_type)

    owners_list = [
        OwnerClass(
            owner=owner_urn,
            type=owner_type,
            source=ownership_source_type,
        )
        for owner_urn in owner_urns
    ]
    return OwnershipClass(
        owners=owners_list,
    )


def make_glossary_terms_aspect_from_urn_list(term_urns: List[str]) -> GlossaryTerms:
    for term_urn in term_urns:
        assert term_urn.startswith("urn:li:glossaryTerm:")
    glossary_terms = GlossaryTerms(
        [GlossaryTermAssociationClass(term_urn) for term_urn in term_urns],
        AuditStampClass(
            time=int(time.time() * 1000),
            actor="urn:li:corpuser:datahub",
        ),
    )
    return glossary_terms


def set_aspect(
    mce: MetadataChangeEventClass, aspect: Optional[Aspect], aspect_type: Type[Aspect]
) -> None:
    """Sets the aspect to the provided aspect, overwriting any previous aspect value that might have existed before.
    If passed in aspect is None, then the existing aspect value will be removed"""
    remove_aspect_if_available(mce, aspect_type)
    if aspect is not None:
        mce.proposedSnapshot.aspects.append(aspect)  # type: ignore



from __future__ import annotations

from datetime import datetime
from typing import Dict, Optional, Type

from typing_extensions import Self

import datahub.metadata.schema_classes as models
from datahub.emitter.mce_builder import ALL_ENV_TYPES
from datahub.emitter.mcp_builder import (
    _INCLUDE_ENV_IN_CONTAINER_PROPERTIES,
    ContainerKey,
)
from datahub.errors import SdkUsageError
from datahub.metadata.urns import (
    ContainerUrn,
    Urn,
)
from datahub.sdk._entity import Entity
from datahub.sdk._shared import (
    DomainInputType,
    HasContainer,
    HasDomain,
    HasOwnership,
    HasPlatformInstance,
    HasSubtype,
    HasTags,
    HasTerms,
    OwnersInputType,
    TagsInputType,
    TermsInputType,
    make_time_stamp,
    parse_time_stamp,
)


class Container(
    HasPlatformInstance,
    HasSubtype,
    HasContainer,
    HasOwnership,
    HasTags,
    HasTerms,
    HasDomain,
    Entity,
):
    __slots__ = ()

    @classmethod
    def get_urn_type(cls) -> Type[ContainerUrn]:
        return ContainerUrn

    def __init__(
        self,
        /,
        # Identity.
        container_key: ContainerKey | ContainerUrn,
        *,
        # Container attributes.
        display_name: str,
        qualified_name: Optional[str] = None,
        description: Optional[str] = None,
        external_url: Optional[str] = None,
        # TODO: call this custom properties?
        extra_properties: Optional[Dict[str, str]] = None,
        created: Optional[datetime] = None,
        last_modified: Optional[datetime] = None,
        # Standard aspects.
        subtype: Optional[str] = None,
        owners: Optional[OwnersInputType] = None,
        tags: Optional[TagsInputType] = None,
        terms: Optional[TermsInputType] = None,
        domain: Optional[DomainInputType] = None,
    ):
        if isinstance(container_key, ContainerUrn):
            urn = container_key
        else:
            urn = ContainerUrn.from_string(container_key.as_urn())
        super().__init__(urn)

        # This needs to come first to ensure that the display name is registered.
        self._ensure_container_props(name=display_name)

        # TODO: Normal usages should require container key. Only the graph init method can accept an urn.
        if isinstance(container_key, ContainerKey):
            self._set_platform_instance(container_key.platform, container_key.instance)

            self._set_container(container_key.parent_key())

            self.set_custom_properties(
                {
                    **container_key.property_dict(),
                    **(extra_properties or {}),
                }
            )

            # Extra validation on the env field.
            # In certain cases (mainly for backwards compatibility), the env field will actually
            # have a platform instance name.
            env = container_key.env if container_key.env in ALL_ENV_TYPES else None
            if _INCLUDE_ENV_IN_CONTAINER_PROPERTIES and env is not None:
                self._ensure_container_props().env = env

        if description is not None:
            self.set_description(description)
        if external_url is not None:
            self.set_external_url(external_url)
        if qualified_name is not None:
            self.set_qualified_name(qualified_name)
        if created is not None:
            self.set_created(created)
        if last_modified is not None:
            self.set_last_modified(last_modified)

        if subtype is not None:
            self.set_subtype(subtype)
        if owners is not None:
            self.set_owners(owners)
        if tags is not None:
            self.set_tags(tags)
        if terms is not None:
            self.set_terms(terms)
        if domain is not None:
            self.set_domain(domain)

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        assert isinstance(urn, ContainerUrn)
        entity = cls(urn, display_name="__dummy_value__")
        return entity._init_from_graph(current_aspects)

    def _ensure_container_props(
        self, *, name: Optional[str] = None
    ) -> models.ContainerPropertiesClass:
        # TODO: Not super happy with this method's implementation, but it's
        # internal-only and enforces the constraints that we need.
        if name is not None:
            return self._setdefault_aspect(models.ContainerPropertiesClass(name=name))

        props = self._get_aspect(models.ContainerPropertiesClass)
        if props is None:
            raise SdkUsageError("Containers must have a name.")
        return props

    @property
    def display_name(self) -> str:
        return self._ensure_container_props().name

    def set_display_name(self, value: str) -> None:
        self._ensure_container_props(name=value).name = value

    @property
    def description(self) -> Optional[str]:
        return self._ensure_container_props().description

    def set_description(self, description: str) -> None:
        self._ensure_container_props().description = description

    @property
    def custom_properties(self) -> Optional[Dict[str, str]]:
        return self._ensure_container_props().customProperties

    def set_custom_properties(self, custom_properties: Dict[str, str]) -> None:
        # TODO: How do we ensure that the container key props are always retained?
        self._ensure_container_props().customProperties = custom_properties

    @property
    def external_url(self) -> Optional[str]:
        return self._ensure_container_props().externalUrl

    def set_external_url(self, external_url: str) -> None:
        self._ensure_container_props().externalUrl = external_url

    @property
    def qualified_name(self) -> Optional[str]:
        return self._ensure_container_props().qualifiedName

    def set_qualified_name(self, qualified_name: str) -> None:
        self._ensure_container_props().qualifiedName = qualified_name

    @property
    def created(self) -> Optional[datetime]:
        return parse_time_stamp(self._ensure_container_props().created)

    def set_created(self, created: datetime) -> None:
        self._ensure_container_props().created = make_time_stamp(created)

    @property
    def last_modified(self) -> Optional[datetime]:
        return parse_time_stamp(self._ensure_container_props().lastModified)

    def set_last_modified(self, last_modified: datetime) -> None:
        self._ensure_container_props().lastModified = make_time_stamp(last_modified)



from typing import ClassVar

from avrogen.dict_wrapper import DictWrapper


class _Aspect(DictWrapper):
    """Base class for all aspects types.

    All codegened types inherit from DictWrapper, either directly or indirectly.
    Types that are aspects inherit directly from _Aspect.
    """

    ASPECT_NAME: ClassVar[str] = None  # type: ignore
    ASPECT_TYPE: ClassVar[str] = "default"
    ASPECT_INFO: ClassVar[dict] = None  # type: ignore

    def __init__(self):
        if type(self) is _Aspect:
            # Ensure that it cannot be instantiated directly, as
            # per https://stackoverflow.com/a/7989101/5004662.
            raise TypeError(
                "_Aspect is an abstract class, and cannot be instantiated directly."
            )
        super().__init__()

    @classmethod
    def get_aspect_name(cls) -> str:
        return cls.ASPECT_NAME  # type: ignore

    @classmethod
    def get_aspect_type(cls) -> str:
        return cls.ASPECT_TYPE

    @classmethod
    def get_aspect_info(cls) -> dict:
        return cls.ASPECT_INFO



import logging
from datetime import datetime
from json import JSONDecodeError
from typing import Dict, List, Literal, Optional, Tuple
from urllib.parse import urlparse

import dateutil.parser
import requests
from pydantic import Field, root_validator

from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    SourceCapability,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.source.dbt.dbt_common import (
    DBTColumn,
    DBTCommonConfig,
    DBTNode,
    DBTSourceBase,
)
from datahub.ingestion.source.dbt.dbt_tests import DBTTest, DBTTestResult

logger = logging.getLogger(__name__)


class DBTCloudConfig(DBTCommonConfig):
    access_url: str = Field(
        description="The base URL of the dbt Cloud instance to use. This should be the URL you use to access the dbt Cloud UI. It should include the scheme (http/https) and not include a trailing slash. See the access url for your dbt Cloud region here: https://docs.getdbt.com/docs/cloud/about-cloud/regions-ip-addresses",
        default="https://cloud.getdbt.com",
    )

    metadata_endpoint: str = Field(
        default="https://metadata.cloud.getdbt.com/graphql",
        description="The dbt Cloud metadata API endpoint. If not provided, we will try to infer it from the access_url.",
    )

    token: str = Field(
        description="The API token to use to authenticate with DBT Cloud.",
    )

    account_id: int = Field(
        description="The DBT Cloud account ID to use.",
    )
    project_id: int = Field(
        description="The dbt Cloud project ID to use.",
    )

    job_id: int = Field(
        description="The ID of the job to ingest metadata from.",
    )
    run_id: Optional[int] = Field(
        None,
        description="The ID of the run to ingest metadata from. If not specified, we'll default to the latest run.",
    )

    external_url_mode: Literal["explore", "ide"] = Field(
        default="explore",
        description='Where should the "View in dbt" link point to - either the "Explore" UI or the dbt Cloud IDE',
    )

    @root_validator(pre=True)
    def set_metadata_endpoint(cls, values: dict) -> dict:
        if values.get("access_url") and not values.get("metadata_endpoint"):
            metadata_endpoint = infer_metadata_endpoint(values["access_url"])
            if metadata_endpoint is None:
                raise ValueError(
                    "Unable to infer the metadata endpoint from the access URL. Please provide a metadata endpoint."
                )
            logger.info(f"Inferred metadata endpoint: {metadata_endpoint}")
            values["metadata_endpoint"] = metadata_endpoint
        return values


def infer_metadata_endpoint(access_url: str) -> Optional[str]:
    """Infer the dbt metadata endpoint from the access URL.

    See https://docs.getdbt.com/docs/cloud/about-cloud/access-regions-ip-addresses#api-access-urls
    and https://docs.getdbt.com/docs/dbt-cloud-apis/discovery-querying#discovery-api-endpoints
    for more information.

    Args:
        access_url: The dbt Cloud access URL. This is the URL of the dbt Cloud UI.

    Returns:
        The metadata endpoint, or None if it couldn't be inferred.

    Examples:
        # Standard multi-tenant deployments.
        >>> infer_metadata_endpoint("https://cloud.getdbt.com")
        'https://metadata.cloud.getdbt.com/graphql'

        >>> infer_metadata_endpoint("https://au.dbt.com")
        'https://metadata.au.dbt.com/graphql'

        >>> infer_metadata_endpoint("https://emea.dbt.com")
        'https://metadata.emea.dbt.com/graphql'

        # Cell-based deployment.
        >>> infer_metadata_endpoint("https://prefix.us1.dbt.com")
        'https://prefix.metadata.us1.dbt.com/graphql'

        # Test with an "internal" URL.
        >>> infer_metadata_endpoint("http://dbt.corp.internal")
        'http://metadata.dbt.corp.internal/graphql'
    """

    try:
        parsed_uri = urlparse(access_url)
        assert parsed_uri.scheme is not None
        assert parsed_uri.hostname is not None
    except Exception as e:
        logger.debug(f"Unable to parse access URL {access_url}: {e}", exc_info=e)
        return None

    if parsed_uri.hostname.endswith(".getdbt.com") or parsed_uri.hostname in {
        # Two special cases of multi-tenant deployments that use the dbt.com domain
        # instead of getdbt.com.
        "au.dbt.com",
        "emea.dbt.com",
    }:
        return f"{parsed_uri.scheme}://metadata.{parsed_uri.netloc}/graphql"
    elif parsed_uri.hostname.endswith(".dbt.com"):
        # For cell-based deployments.
        # prefix.region.dbt.com -> prefix.metadata.region.dbt.com
        hostname_parts = parsed_uri.hostname.split(".", maxsplit=1)
        return f"{parsed_uri.scheme}://{hostname_parts[0]}.metadata.{hostname_parts[1]}/graphql"
    else:
        # The self-hosted variants also have the metadata. prefix.
        return f"{parsed_uri.scheme}://metadata.{parsed_uri.netloc}/graphql"


_DBT_GRAPHQL_COMMON_FIELDS = """
  runId
  accountId
  projectId
  environmentId
  jobId
  resourceType
  uniqueId
  name
  description
  meta
  dbtVersion
  tags
"""

_DBT_GRAPHQL_NODE_COMMON_FIELDS = """
  database
  schema
  type
  owner
  comment

  columns {
    name
    index
    type
    comment
    description
    tags
    meta
  }

  # We don't currently support this field, but should in the future.
  #stats {
  #  id
  #  label
  #  description
  #  include
  #  value
  #}
"""

_DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS = """
  packageName
  alias
  error
  status
  skip
  rawSql
  rawCode
  compiledSql
  compiledCode
"""

_DBT_FIELDS_BY_TYPE = {
    "models": f"""
    {_DBT_GRAPHQL_COMMON_FIELDS}
    {_DBT_GRAPHQL_NODE_COMMON_FIELDS}
    {_DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS}
    dependsOn
    materializedType
""",
    "seeds": f"""
    {_DBT_GRAPHQL_COMMON_FIELDS}
    {_DBT_GRAPHQL_NODE_COMMON_FIELDS}
    {_DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS}
""",
    "sources": f"""
    {_DBT_GRAPHQL_COMMON_FIELDS}
    {_DBT_GRAPHQL_NODE_COMMON_FIELDS}
    identifier
    sourceName
    sourceDescription
    maxLoadedAt
    snapshottedAt
    state
    freshnessChecked
    loader
""",
    "snapshots": f"""
    {_DBT_GRAPHQL_COMMON_FIELDS}
    {_DBT_GRAPHQL_NODE_COMMON_FIELDS}
    {_DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS}
    parentsSources {{
      uniqueId
    }}
    parentsModels {{
      uniqueId
    }}
""",
    "tests": f"""
    {_DBT_GRAPHQL_COMMON_FIELDS}
    state
    columnName
    status
    error
    dependsOn
    fail
    warn
    skip
    rawSql
    rawCode
    compiledSql
    compiledCode
""",
    # Currently unsupported dbt node types:
    # - metrics
    # - snapshots
    # - exposures
}

_DBT_GRAPHQL_QUERY = """
query DatahubMetadataQuery_{type}($jobId: BigInt!, $runId: BigInt) {{
  job(id: $jobId, runId: $runId) {{
    {type} {{
{fields}
    }}
  }}
}}
"""


@platform_name("dbt")
@config_class(DBTCloudConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.DELETION_DETECTION, "Enabled via stateful ingestion")
@capability(SourceCapability.LINEAGE_COARSE, "Enabled by default")
class DBTCloudSource(DBTSourceBase, TestableSource):
    config: DBTCloudConfig

    @classmethod
    def create(cls, config_dict, ctx):
        config = DBTCloudConfig.parse_obj(config_dict)
        return cls(config, ctx, "dbt")

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        test_report = TestConnectionReport()
        try:
            source_config = DBTCloudConfig.parse_obj_allow_extras(config_dict)
            DBTCloudSource._send_graphql_query(
                metadata_endpoint=source_config.metadata_endpoint,
                token=source_config.token,
                query=_DBT_GRAPHQL_QUERY.format(type="tests", fields="jobId"),
                variables={
                    "jobId": source_config.job_id,
                    "runId": source_config.run_id,
                },
            )
            test_report.basic_connectivity = CapabilityReport(capable=True)
        except Exception as e:
            test_report.basic_connectivity = CapabilityReport(
                capable=False, failure_reason=str(e)
            )
        return test_report

    @staticmethod
    def _send_graphql_query(
        metadata_endpoint: str, token: str, query: str, variables: Dict
    ) -> Dict:
        logger.debug(f"Sending GraphQL query to dbt Cloud: {query}")
        response = requests.post(
            metadata_endpoint,
            json={
                "query": query,
                "variables": variables,
            },
            headers={
                "Authorization": f"Bearer {token}",
                "X-dbt-partner-source": "acryldatahub",
            },
        )

        try:
            res = response.json()
            if "errors" in res:
                raise ValueError(
                    f"Unable to fetch metadata from dbt Cloud: {res['errors']}"
                )
            data = res["data"]
        except JSONDecodeError as e:
            response.raise_for_status()
            raise e

        return data

    def load_nodes(self) -> Tuple[List[DBTNode], Dict[str, Optional[str]]]:
        # TODO: In dbt Cloud, commands are scheduled as part of jobs, where
        # each job can have multiple runs. We currently only fully support
        # jobs that do a full / mostly full build of the project, and will
        # print out warnings if data is missing. In the future, it'd be nice
        # to automatically detect all jobs that are part of a project and combine
        # their metadata together.
        # Additionally, we'd like to model dbt Cloud jobs/runs in DataHub
        # as DataProcesses or DataJobs.

        raw_nodes = []
        for node_type, fields in _DBT_FIELDS_BY_TYPE.items():
            logger.info(f"Fetching {node_type} from dbt Cloud")
            data = self._send_graphql_query(
                metadata_endpoint=self.config.metadata_endpoint,
                token=self.config.token,
                query=_DBT_GRAPHQL_QUERY.format(type=node_type, fields=fields),
                variables={
                    "jobId": self.config.job_id,
                    "runId": self.config.run_id,
                },
            )

            raw_nodes.extend(data["job"][node_type])

        nodes = [self._parse_into_dbt_node(node) for node in raw_nodes]

        additional_metadata: Dict[str, Optional[str]] = {
            "project_id": str(self.config.project_id),
            "account_id": str(self.config.account_id),
            "job_id": str(self.config.job_id),
        }

        return nodes, additional_metadata

    def _parse_into_dbt_node(self, node: Dict) -> DBTNode:
        key = node["uniqueId"]

        name = node["name"]
        if self.config.use_identifiers and node.get("identifier"):
            name = node["identifier"]
        if node["resourceType"] != "test" and node.get("alias"):
            name = node["alias"]

        comment = node.get("comment", "")
        description = node["description"]
        if node.get("sourceDescription"):
            description = node["sourceDescription"]

        if node["resourceType"] == "model":
            materialization = node["materializedType"]
        elif node["resourceType"] == "snapshot":
            materialization = "snapshot"
        else:
            materialization = None

        if node["resourceType"] == "snapshot":
            upstream_nodes = [
                obj["uniqueId"]
                for obj in [
                    *node.get("parentsModels", []),
                    *node.get("parentsSources", []),
                ]
            ]
        else:
            upstream_nodes = node.get("dependsOn", [])

        catalog_type = node.get("type")

        meta = node["meta"]

        # The dbt owner field is set to the db user, and not the meta property.
        # owner = node.get("owner")
        owner = meta.get("owner")

        tags = node["tags"]
        tags = [self.config.tag_prefix + tag for tag in tags]

        if node["resourceType"] in {"model", "seed", "snapshot"}:
            status = node["status"]
            if status is None and materialization != "ephemeral":
                self.report.report_warning(
                    key, "node is missing a status, schema metadata will be incomplete"
                )

            # The code fields are new in dbt 1.3, and replace the sql ones.
            raw_code = node["rawCode"] or node["rawSql"]
            compiled_code = node["compiledCode"] or node["compiledSql"]
        else:
            raw_code = None
            compiled_code = None

        max_loaded_at = None
        if node["resourceType"] == "source":
            max_loaded_at_str = node["maxLoadedAt"]
            if max_loaded_at_str:
                max_loaded_at = dateutil.parser.parse(max_loaded_at_str)

                # For missing data, dbt returns 0001-01-01T00:00:00.000Z.
                if max_loaded_at.year <= 1:
                    max_loaded_at = None

        columns = []
        if "columns" in node and node["columns"] is not None:
            # columns will be empty for ephemeral models
            columns = list(
                sorted(
                    [self._parse_into_dbt_column(column) for column in node["columns"]],
                    key=lambda c: c.index,
                )
            )

        test_info = None
        test_result = None
        if node["resourceType"] == "test":
            qualified_test_name = name

            # The qualified test name should be the test name from the dbt project.
            # It can be simple (e.g. 'unique') or prefixed (e.g. 'dbt_expectations.expect_column_values_to_not_be_null').
            # We attempt to guess the test name based on the macros used.
            for dependency in node["dependsOn"]:
                # An example dependsOn list could be:
                #     ['model.sample_dbt.monthly_billing_with_cust', 'macro.dbt.test_not_null', 'macro.dbt.get_where_subquery']
                # In that case, the test should be `not_null`.

                if dependency.startswith("macro."):
                    _, macro = dependency.split(".", 1)
                    if macro.startswith("dbt."):
                        if not macro.startswith("dbt.test_"):
                            continue
                        macro = macro[len("dbt.test_") :]

                    qualified_test_name = macro
                    break

            test_info = DBTTest(
                qualified_test_name=qualified_test_name,
                column_name=node["columnName"],
                kw_args={},  # TODO: dbt Cloud doesn't expose the args.
            )
            if not node["skip"]:
                test_result = DBTTestResult(
                    invocation_id=f"job{node['jobId']}-run{node['runId']}",
                    execution_time=datetime.now(),  # TODO: dbt Cloud doesn't expose this.
                    status=node["status"],
                    native_results={
                        key: str(node[key])
                        for key in {
                            "columnName",
                            "error",
                            "fail",
                            "warn",
                            "skip",
                            "state",
                            "status",
                        }
                    },
                )

        return DBTNode(
            dbt_name=key,
            # TODO: Get the dbt adapter natively.
            dbt_adapter=self.config.target_platform,
            dbt_package_name=node.get("packageName"),
            database=node.get("database"),
            schema=node.get("schema"),
            name=name,
            alias=node.get("alias"),
            dbt_file_path=None,  # TODO: Get this from the dbt API.
            node_type=node["resourceType"],
            max_loaded_at=max_loaded_at,
            comment=comment,
            description=description,
            upstream_nodes=upstream_nodes,
            materialization=materialization,
            catalog_type=catalog_type,
            missing_from_catalog=False,  # This doesn't really apply to dbt Cloud.
            meta=meta,
            query_tag={},  # TODO: Get this from the dbt API.
            tags=tags,
            owner=owner,
            language="sql",  # TODO: dbt Cloud doesn't surface this
            raw_code=raw_code,
            compiled_code=compiled_code,
            columns=columns,
            test_info=test_info,
            test_results=[test_result] if test_result else [],
            model_performances=[],  # TODO: support model performance with dbt Cloud
        )

    def _parse_into_dbt_column(
        self,
        column: Dict,
    ) -> DBTColumn:
        return DBTColumn(
            name=column["name"],
            comment=column.get("comment", ""),
            description=column["description"],
            # For some reason, the index sometimes comes back as None from the dbt Cloud API.
            # In that case, we just assume that the column is at the end of the table by
            # assigning it a very large index.
            index=column["index"] if column["index"] is not None else 10**6,
            data_type=column["type"],
            meta=column["meta"],
            tags=column["tags"],
        )

    def get_external_url(self, node: DBTNode) -> Optional[str]:
        if self.config.external_url_mode == "explore":
            return f"{self.config.access_url}/explore/{self.config.account_id}/projects/{self.config.project_id}/environments/production/details/{node.dbt_name}"
        else:
            return f"{self.config.access_url}/develop/{self.config.account_id}/projects/{self.config.project_id}"



import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from databricks.sdk.service.catalog import DataSourceFormat
from sqlalchemy import create_engine
from sqlalchemy.engine import Connection

from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.sql.sql_config import SQLCommonConfig
from datahub.ingestion.source.sql.sql_generic import BaseTable
from datahub.ingestion.source.sql.sql_generic_profiler import (
    GenericProfiler,
    TableProfilerRequest,
)
from datahub.ingestion.source.unity.config import UnityCatalogGEProfilerConfig
from datahub.ingestion.source.unity.proxy_types import Table, TableReference
from datahub.ingestion.source.unity.report import UnityCatalogReport

logger = logging.getLogger(__name__)


@dataclass(init=False)
class UnityCatalogSQLGenericTable(BaseTable):
    ref: TableReference = field(init=False)

    def __init__(self, table: Table):
        self.name = table.name
        self.comment = table.comment
        self.created = table.created_at
        self.last_altered = table.updated_at
        self.column_count = len(table.columns)
        self.ref = table.ref
        self.size_in_bytes = None
        self.rows_count = None
        self.ddl = None
        self.data_source_format = table.data_source_format

    @property
    def is_delta_table(self) -> bool:
        return self.data_source_format == DataSourceFormat.DELTA


class UnityCatalogGEProfiler(GenericProfiler):
    sql_common_config: SQLCommonConfig
    profiling_config: UnityCatalogGEProfilerConfig
    report: UnityCatalogReport

    def __init__(
        self,
        sql_common_config: SQLCommonConfig,
        profiling_config: UnityCatalogGEProfilerConfig,
        report: UnityCatalogReport,
    ) -> None:
        super().__init__(sql_common_config, report, "databricks")
        self.profiling_config = profiling_config
        # TODO: Consider passing dataset urn builder directly
        # So there is no repeated logic between this class and source.py

    def get_workunits(self, tables: List[Table]) -> Iterable[MetadataWorkUnit]:
        # Extra default SQLAlchemy option for better connection pooling and threading.
        # https://docs.sqlalchemy.org/en/14/core/pooling.html#sqlalchemy.pool.QueuePool.params.max_overflow
        self.config.options.setdefault(
            "max_overflow", self.profiling_config.max_workers
        )

        url = self.config.get_sql_alchemy_url()
        engine = create_engine(url, **self.config.options)
        conn = engine.connect()

        profile_requests = []
        with ThreadPoolExecutor(
            max_workers=self.profiling_config.max_workers
        ) as executor:
            futures = [
                executor.submit(
                    self.get_unity_profile_request,
                    UnityCatalogSQLGenericTable(table),
                    conn,
                )
                for table in tables
            ]

            try:
                for i, completed in enumerate(
                    as_completed(futures, timeout=self.profiling_config.max_wait_secs)
                ):
                    profile_request = completed.result()
                    if profile_request is not None:
                        profile_requests.append(profile_request)
                    if i > 0 and i % 100 == 0:
                        logger.info(f"Finished table-level profiling for {i} tables")
            except TimeoutError:
                logger.warning("Timed out waiting to complete table-level profiling.")

        if len(profile_requests) == 0:
            return

        yield from self.generate_profile_workunits(
            profile_requests,
            max_workers=self.config.profiling.max_workers,
            platform=self.platform,
            profiler_args=self.get_profile_args(),
        )

    def get_dataset_name(self, table_name: str, schema_name: str, db_name: str) -> str:
        # Note: unused... ideally should share logic with TableReference
        return f"{db_name}.{schema_name}.{table_name}"

    def get_unity_profile_request(
        self, table: UnityCatalogSQLGenericTable, conn: Connection
    ) -> Optional[TableProfilerRequest]:
        # TODO: Reduce code duplication with get_profile_request
        skip_profiling = False
        profile_table_level_only = self.profiling_config.profile_table_level_only

        dataset_name = table.ref.qualified_table_name
        if table.is_delta_table:
            try:
                table.size_in_bytes = _get_dataset_size_in_bytes(table, conn)
            except Exception as e:
                self.report.warning(
                    title="Incomplete Dataset Profile",
                    message="Failed to get table size",
                    context=dataset_name,
                    exc=e,
                )

        if table.size_in_bytes is None:
            self.report.num_profile_missing_size_in_bytes += 1

        if not self.is_dataset_eligible_for_profiling(
            dataset_name,
            size_in_bytes=table.size_in_bytes,
            last_altered=table.last_altered,
            rows_count=0,  # Can't get row count ahead of time
        ):
            # Profile only table level if dataset is filtered from profiling
            # due to size limits alone
            if self.is_dataset_eligible_for_profiling(
                dataset_name,
                last_altered=table.last_altered,
                size_in_bytes=None,
                rows_count=None,
            ):
                profile_table_level_only = True
            else:
                skip_profiling = True

        if table.column_count == 0:
            skip_profiling = True

        if skip_profiling:
            if self.profiling_config.report_dropped_profiles:
                self.report.report_dropped(dataset_name)
            return None

        if profile_table_level_only and table.is_delta_table:
            # For requests with profile_table_level_only set, dataset profile is generated
            # by looking at table.rows_count. For delta tables (a typical databricks table)
            # count(*) is an efficient query to compute row count.
            try:
                table.rows_count = _get_dataset_row_count(table, conn)
            except Exception as e:
                self.report.warning(
                    title="Incomplete Dataset Profile",
                    message="Failed to get table row count",
                    context=dataset_name,
                    exc=e,
                )

        if table.rows_count is None:
            self.report.num_profile_missing_row_count += 1

        self.report.report_entity_profiled(dataset_name)
        logger.debug(f"Preparing profiling request for {dataset_name}")
        return TableProfilerRequest(
            table=table,
            pretty_name=dataset_name,
            batch_kwargs=dict(schema=table.ref.schema, table=table.name),
            profile_table_level_only=profile_table_level_only,
        )


def _get_dataset_size_in_bytes(
    table: UnityCatalogSQLGenericTable, conn: Connection
) -> Optional[int]:
    name = ".".join(
        conn.dialect.identifier_preparer.quote(c)
        for c in [table.ref.catalog, table.ref.schema, table.ref.table]
    )
    # This query only works for delta table.
    # Ref: https://docs.databricks.com/en/delta/table-details.html
    # Note: Any change here should also update _get_dataset_row_count
    row = conn.execute(f"DESCRIBE DETAIL {name}").fetchone()
    if row is None:
        return None
    else:
        try:
            return int(row._asdict()["sizeInBytes"])
        except Exception:
            return None


def _get_dataset_row_count(
    table: UnityCatalogSQLGenericTable, conn: Connection
) -> Optional[int]:
    name = ".".join(
        conn.dialect.identifier_preparer.quote(c)
        for c in [table.ref.catalog, table.ref.schema, table.ref.table]
    )
    # This query only works efficiently for delta table
    row = conn.execute(f"select count(*) as numRows from {name}").fetchone()
    if row is None:
        return None
    else:
        try:
            return int(row._asdict()["numRows"])
        except Exception:
            return None



import logging
from functools import partial
from typing import Any, Iterable, List, Optional, Union

import progressbar
from pydantic import Field

from datahub.configuration.common import ConfigModel
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, Source, SourceReport
from datahub.ingestion.api.source_helpers import auto_workunit_reporter
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.metadata.schema_classes import (
    DomainsClass,
    GlossaryTermAssociationClass,
    MetadataChangeProposalClass,
    OwnerClass,
    OwnershipTypeClass,
    TagAssociationClass,
)
from datahub.specific.dataset import DatasetPatchBuilder

logger = logging.getLogger(__name__)


def apply_association_to_container(
    container_urn: str,
    association_urn: str,
    association_type: str,
    emit: bool = True,
    graph: Optional[DataHubGraph] = None,
) -> Optional[List[Union[MetadataChangeProposalWrapper, MetadataChangeProposalClass]]]:
    """
    Common function to add either tags, terms, domains, or owners to child datasets (for now).

    Args:
        container_urn: The URN of the container
        association_urn: The URN of the tag, term, or user to apply
        association_type: One of 'tag', 'term', 'domain' or 'owner'
    """
    urns: List[str] = [container_urn]
    if not graph:
        graph = get_default_graph()
    logger.info(f"Using {graph}")
    urns.extend(
        graph.get_urns_by_filter(
            container=container_urn,
            batch_size=1000,
            entity_types=["dataset", "container"],
        )
    )

    all_patches: List[Any] = []
    for urn in urns:
        builder = DatasetPatchBuilder(urn)
        patches: List[Any] = []
        if association_type == "tag":
            patches = builder.add_tag(TagAssociationClass(association_urn)).build()
        elif association_type == "term":
            patches = builder.add_term(
                GlossaryTermAssociationClass(association_urn)
            ).build()
        elif association_type == "owner":
            patches = builder.add_owner(
                OwnerClass(
                    owner=association_urn,
                    type=OwnershipTypeClass.TECHNICAL_OWNER,
                )
            ).build()
        elif association_type == "domain":
            patches = [
                MetadataChangeProposalWrapper(
                    entityUrn=urn,
                    aspect=DomainsClass(domains=[association_urn]),
                )
            ]
        all_patches.extend(patches)
    if emit:
        mcps_iter = progressbar.progressbar(all_patches, redirect_stdout=True)
        for mcp in mcps_iter:
            graph.emit(mcp)
        return None
    else:
        return all_patches


class DomainApplyConfig(ConfigModel):
    assets: List[str] = Field(
        default_factory=list,
        description="List of assets to apply domain hierarchichaly. Currently only containers and datasets are supported",
    )
    domain_urn: str = Field(default="")


class TagApplyConfig(ConfigModel):
    assets: List[str] = Field(
        default_factory=list,
        description="List of assets to apply tag hierarchichaly. Currently only containers and datasets are supported",
    )
    tag_urn: str = Field(default="")


class TermApplyConfig(ConfigModel):
    assets: List[str] = Field(
        default_factory=list,
        description="List of assets to apply term hierarchichaly. Currently only containers and datasets are supported",
    )
    term_urn: str = Field(default="")


class OwnerApplyConfig(ConfigModel):
    assets: List[str] = Field(
        default_factory=list,
        description="List of assets to apply owner hierarchichaly. Currently only containers and datasets are supported",
    )
    owner_urn: str = Field(default="")


class DataHubApplyConfig(ConfigModel):
    domain_apply: Optional[List[DomainApplyConfig]] = Field(
        default=None,
        description="List to apply domains to assets",
    )
    tag_apply: Optional[List[TagApplyConfig]] = Field(
        default=None,
        description="List to apply tags to assets",
    )
    term_apply: Optional[List[TermApplyConfig]] = Field(
        default=None,
        description="List to apply terms to assets",
    )
    owner_apply: Optional[List[OwnerApplyConfig]] = Field(
        default=None,
        description="List to apply owners to assets",
    )


@platform_name("DataHubApply")
@config_class(DataHubApplyConfig)
@support_status(SupportStatus.TESTING)
class DataHubApplySource(Source):
    """
    This source is a helper over CLI
    so people can use the helper to apply various metadata changes to DataHub
    via Managed Ingestion
    """

    def __init__(self, ctx: PipelineContext, config: DataHubApplyConfig):
        self.ctx = ctx
        self.config = config
        self.report = SourceReport()
        self.graph = ctx.require_graph()

    def _yield_workunits(
        self,
        proposals: List[
            Union[MetadataChangeProposalWrapper, MetadataChangeProposalClass]
        ],
    ) -> Iterable[MetadataWorkUnit]:
        for proposal in proposals:
            if isinstance(proposal, MetadataChangeProposalWrapper):
                yield proposal.as_workunit()
            else:
                yield MetadataWorkUnit(
                    id=MetadataWorkUnit.generate_workunit_id(proposal),
                    mcp_raw=proposal,
                )

    def _handle_assets(
        self, assets: List[str], apply_urn: str, apply_type: str
    ) -> Iterable[MetadataWorkUnit]:
        for asset in assets:
            change_proposals = apply_association_to_container(
                asset, apply_urn, apply_type, emit=False, graph=self.graph
            )
            assert change_proposals is not None
            yield from self._yield_workunits(change_proposals)

    def _yield_domain(self) -> Iterable[MetadataWorkUnit]:
        if not self.config.domain_apply:
            return
        for apply in self.config.domain_apply:
            yield from self._handle_assets(apply.assets, apply.domain_urn, "domain")

    def _yield_tag(self) -> Iterable[MetadataWorkUnit]:
        if not self.config.tag_apply:
            return
        for apply in self.config.tag_apply:
            yield from self._handle_assets(apply.assets, apply.tag_urn, "tag")

    def _yield_term(self) -> Iterable[MetadataWorkUnit]:
        if not self.config.term_apply:
            return
        for apply in self.config.term_apply:
            yield from self._handle_assets(apply.assets, apply.term_urn, "term")

    def _yield_owner(self) -> Iterable[MetadataWorkUnit]:
        if not self.config.owner_apply:
            return
        for apply in self.config.owner_apply:
            yield from self._handle_assets(apply.assets, apply.owner_urn, "owner")

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        yield from self._yield_domain()
        yield from self._yield_tag()
        yield from self._yield_term()
        yield from self._yield_owner()

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [partial(auto_workunit_reporter, self.get_report())]

    def get_report(self) -> SourceReport:
        return self.report



import dataclasses
import json
import re
from collections import defaultdict
from typing import Any, Dict, List, Sequence, Set, Tuple, Union

import deepdiff.serialization
import yaml
from deepdiff import DeepDiff
from deepdiff.model import DiffLevel
from deepdiff.operator import BaseOperator
from typing_extensions import Literal

ReportType = Literal[
    "type_changes",
    "dictionary_item_added",
    "dictionary_item_removed",
    "values_changed",
    "unprocessed",
    "iterable_item_added",
    "iterable_item_removed",
    "iterable_item_moved",
    "attribute_added",
    "attribute_removed",
    "set_item_added",
    "set_item_removed",
    "repetition_change",
]


@dataclasses.dataclass(frozen=True)
class AspectForDiff:
    urn: str
    change_type: str
    aspect_name: str
    aspect: Dict[str, Any] = dataclasses.field(hash=False)
    delta_info: "DeltaInfo" = dataclasses.field(hash=False, repr=False)

    @classmethod
    def create_from_mcp(cls, idx: int, obj: Dict[str, Any]) -> "AspectForDiff":
        aspect = obj["aspect"]
        return cls(
            urn=obj["entityUrn"],
            change_type=obj["changeType"],
            aspect_name=obj["aspectName"],
            aspect=aspect.get("json", aspect),
            delta_info=DeltaInfo(idx=idx, original=obj),
        )

    def __repr__(self):
        d = {
            field.name: getattr(self, field.name)
            for field in dataclasses.fields(self)
            if field.repr
        }
        d["aspect"] = "<aspect>"
        return "\n" + "\t" * 2 + str(json.dumps(d))


@dataclasses.dataclass
class DeltaInfo:
    """Information about an MCP used to construct a diff delta.

    In a separate class so it can be ignored by DeepDiff via MCPDeltaInfoOperator.
    """

    idx: int  # Location in list of MCEs in golden file
    original: Dict[str, Any]  # Original json-serialized MCP


class DeltaInfoOperator(BaseOperator):
    """Warning: Doesn't seem to be working right now.
    Ignored via an ignore path as an extra layer of defense.
    """

    def __init__(self):
        super().__init__(types=[DeltaInfo])

    def give_up_diffing(self, *args: Any, **kwargs: Any) -> bool:
        return True


AspectsByUrn = Dict[str, Dict[str, List[AspectForDiff]]]


class CannotCompareMCPs(Exception):
    pass


def get_aspects_by_urn(obj: object) -> AspectsByUrn:
    """Restructure a list of serialized MCPs by urn and aspect.
    Retains information like the original dict and index to facilitate `apply_delta` later.

    Raises:
        AssertionError: If the input is not purely a list of MCPs.
    """
    d: AspectsByUrn = defaultdict(dict)
    assert isinstance(obj, list), obj
    for i, entry in enumerate(obj):
        assert isinstance(entry, dict), entry
        if "proposedSnapshot" in entry:
            raise CannotCompareMCPs("Found MCEs")
        elif "entityUrn" in entry and "aspectName" in entry and "aspect" in entry:
            urn = entry["entityUrn"]
            aspect_name = entry["aspectName"]
            aspect = AspectForDiff.create_from_mcp(i, entry)
            d[urn].setdefault(aspect_name, []).append(aspect)
        else:
            raise AssertionError(f"Unrecognized MCE: {entry}")

    return d


@dataclasses.dataclass
class MCPAspectDiff:
    aspects_added: Dict[int, AspectForDiff]
    aspects_removed: Dict[int, AspectForDiff]
    aspects_changed: Dict[Tuple[int, AspectForDiff, AspectForDiff], List[DiffLevel]]

    @classmethod
    def create(cls, diff: DeepDiff) -> "MCPAspectDiff":
        # Parse DeepDiff to distinguish between aspects that were added, removed, or changed
        aspects_added = {}
        aspects_removed = {}
        aspects_changed = defaultdict(list)
        for key, diff_levels in diff.tree.items():
            for diff_level in diff_levels:
                path = diff_level.path(output_format="list")
                idx = int(path[0])
                if len(path) == 1 and key == "iterable_item_added":
                    aspects_added[idx] = diff_level.t2
                elif len(path) == 1 and key == "iterable_item_removed":
                    aspects_removed[idx] = diff_level.t1
                else:
                    level = diff_level
                    while not isinstance(level.t1, AspectForDiff):
                        level = level.up
                    aspects_changed[(idx, level.t1, level.t2)].append(diff_level)

        return cls(
            aspects_added=aspects_added,
            aspects_removed=aspects_removed,
            aspects_changed=aspects_changed,
        )


@dataclasses.dataclass
class MCPDiff:
    aspect_changes: Dict[str, Dict[str, MCPAspectDiff]]  # urn -> aspect -> diff
    urns_added: Set[str]
    urns_removed: Set[str]
    is_delta_valid: bool

    def __bool__(self) -> bool:
        return bool(self.aspect_changes)

    @classmethod
    def create(
        cls,
        golden: AspectsByUrn,
        output: AspectsByUrn,
        ignore_paths: Sequence[str],
    ) -> "MCPDiff":
        ignore_paths = [cls.convert_path(path) for path in ignore_paths]

        is_delta_valid = True
        aspect_changes: Dict[str, Dict[str, MCPAspectDiff]] = defaultdict(dict)
        for urn in golden.keys() | output.keys():
            golden_map = golden.get(urn, {})
            output_map = output.get(urn, {})
            for aspect_name in golden_map.keys() | output_map.keys():
                t1 = golden_map.get(aspect_name, [])
                t2 = output_map.get(aspect_name, [])
                diff = DeepDiff(
                    t1=t1,
                    t2=t2,
                    exclude_regex_paths=ignore_paths,
                    ignore_order=True,
                    custom_operators=[DeltaInfoOperator()],
                )
                if diff:
                    aspect_changes[urn][aspect_name] = MCPAspectDiff.create(diff)
                if len(t1) > 1 or len(t2) > 1:
                    is_delta_valid = False

        return cls(
            urns_added=output.keys() - golden.keys(),
            urns_removed=golden.keys() - output.keys(),
            aspect_changes=aspect_changes,
            is_delta_valid=is_delta_valid,
        )

    @staticmethod
    def convert_path(path: str) -> str:
        # Attempt to use paths intended for the root golden... sorry for the regex
        return re.sub(
            r"^root\\?\[([0-9]+|\\d\+)\\?]\\?\['aspect'\\?](\\?\['(json|value)'\\?])?",
            r"root\[\\d+].aspect",
            path,
        )

    def apply_delta(self, golden: List[Dict[str, Any]]) -> None:
        """Update a golden file to match an output file based on the diff.

        :param golden: Golden file represented as a list of MCPs, altered in-place.
        """
        aspect_diffs = [v for d in self.aspect_changes.values() for v in d.values()]
        for aspect_diff in aspect_diffs:
            for _, old, new in aspect_diff.aspects_changed.keys():
                golden[old.delta_info.idx] = new.delta_info.original

        indices_to_remove = set()
        for aspect_diff in aspect_diffs:
            for ga in aspect_diff.aspects_removed.values():
                indices_to_remove.add(ga.delta_info.idx)
        for idx in sorted(indices_to_remove, reverse=True):
            del golden[idx]

        for aspect_diff in aspect_diffs:  # Ideally would have smarter way to do this
            for ga in aspect_diff.aspects_added.values():
                golden.insert(ga.delta_info.idx, ga.delta_info.original)

    def pretty(self, verbose: bool = False) -> str:
        """The pretty human-readable string output of the diff between golden and output."""
        s = []
        for urn in self.urns_added:
            s.append(f"Urn added, {urn}{' with aspects:' if verbose else ''}")
            if verbose:
                for aspect_diff in self.aspect_changes[urn].values():
                    for i, ga in aspect_diff.aspects_added.items():
                        s.append(self.report_aspect(ga, i))
                        s.append(serialize_aspect(ga.aspect))
        if self.urns_added:
            s.append("")

        for urn in self.urns_removed:
            s.append(f"Urn removed, {urn}{' with aspects:' if verbose else ''}")
            if verbose:
                for aspect_diff in self.aspect_changes[urn].values():
                    for i, ga in aspect_diff.aspects_removed.items():
                        s.append(self.report_aspect(ga, i))
                        s.append(serialize_aspect(ga.aspect))
        if self.urns_removed:
            s.append("")

        for urn in self.aspect_changes.keys() - self.urns_added - self.urns_removed:
            aspect_map = self.aspect_changes[urn]
            s.append(f"Urn changed, {urn}:")
            for aspect_diffs in aspect_map.values():
                for i, ga in aspect_diffs.aspects_added.items():
                    s.append(self.report_aspect(ga, i, "added"))
                    if verbose:
                        s.append(serialize_aspect(ga.aspect))
                for i, ga in aspect_diffs.aspects_removed.items():
                    s.append(self.report_aspect(ga, i, "removed"))
                    if verbose:
                        s.append(serialize_aspect(ga.aspect))
                for (i, old, new), diffs in aspect_diffs.aspects_changed.items():
                    s.append(self.report_aspect(old, i, "changed") + ":")
                    for diff_level in diffs:
                        s.append(self.report_diff_level(diff_level, i))
                    if verbose:
                        s.append(f"Old aspect:\n{serialize_aspect(old.aspect)}")
                        s.append(f"New aspect:\n{serialize_aspect(new.aspect)}")

            s.append("")

        return "\n".join(s)

    @staticmethod
    def report_aspect(ga: AspectForDiff, idx: int, msg: str = "") -> str:
        # Describe as "nth <aspect>" if n > 1
        base = (idx + 1) % 10
        if base == 1:
            suffix = "st"
        elif base == 2:
            suffix = "nd"
        elif base == 3:
            suffix = "rd"
        else:
            suffix = "th"
        ordinal = f"{(idx + 1)}{suffix} " if idx else ""
        return f"{ordinal}<{ga.aspect_name}> {msg}"

    @staticmethod
    def report_diff_level(diff: DiffLevel, idx: int) -> str:
        return "\t" + deepdiff.serialization.pretty_print_diff(diff).replace(
            f"root[{idx}].", ""
        )


def serialize_aspect(aspect: Union[AspectForDiff, Dict[str, Any]]) -> str:
    if isinstance(aspect, AspectForDiff):  # Unpack aspect
        aspect = aspect.aspect
    return "    " + yaml.dump(aspect, sort_keys=False).replace("\n", "\n    ").strip()



import logging
import re
import time
from typing import Any, Dict, Iterable, List, Tuple

from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.dremio.dremio_api import (
    DremioAPIException,
    DremioAPIOperations,
)
from datahub.ingestion.source.dremio.dremio_config import DremioSourceConfig
from datahub.ingestion.source.dremio.dremio_entities import DremioDataset
from datahub.ingestion.source.dremio.dremio_reporting import DremioSourceReport
from datahub.metadata.schema_classes import (
    DatasetFieldProfileClass,
    DatasetProfileClass,
    QuantileClass,
)

logger = logging.getLogger(__name__)


class DremioProfiler:
    config: DremioSourceConfig
    report: DremioSourceReport
    MAX_COLUMNS_PER_QUERY = 800

    def __init__(
        self,
        config: DremioSourceConfig,
        report: DremioSourceReport,
        api_operations: DremioAPIOperations,
    ) -> None:
        self.api_operations = api_operations
        self.config = config
        self.report = report
        self.QUERY_TIMEOUT = (
            config.profiling.query_timeout
        )  # 5 minutes timeout for each query

    def get_workunits(
        self, dataset: DremioDataset, dataset_urn: str
    ) -> Iterable[MetadataWorkUnit]:
        if not dataset.columns:
            self.report.warning(
                message="Skipping profiling as no columns found for table",
                context=f"{dataset.resource_name}",
            )
            self.report.profiling_skipped_other[dataset.resource_name] += 1
            return

        full_table_name = (
            '"' + '"."'.join(dataset.path) + '"."' + dataset.resource_name + '"'
        )

        columns = [(col.name, col.data_type) for col in dataset.columns]

        if not self.config.profile_pattern.allowed(full_table_name):
            self.report.profiling_skipped_table_profile_pattern[full_table_name] += 1
            self.report.warning(
                message="Profiling is restricted due to the specified profile pattern.",
                context=f"{full_table_name}",
            )
            return

        profile_data = self.profile_table(full_table_name, columns)
        profile_aspect = self.populate_profile_aspect(profile_data)

        if profile_aspect:
            self.report.report_entity_profiled(dataset.resource_name)
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn, aspect=profile_aspect
            )
            yield mcp.as_workunit()

    def populate_profile_aspect(self, profile_data: Dict) -> DatasetProfileClass:
        field_profiles = [
            self._create_field_profile(field_name, field_stats)
            for field_name, field_stats in profile_data.get("column_stats", {}).items()
        ]
        return DatasetProfileClass(
            timestampMillis=round(time.time() * 1000),
            rowCount=profile_data.get("row_count"),
            columnCount=profile_data.get("column_count"),
            fieldProfiles=field_profiles,
        )

    def _create_field_profile(
        self, field_name: str, field_stats: Dict
    ) -> DatasetFieldProfileClass:
        quantiles = field_stats.get("quantiles")
        return DatasetFieldProfileClass(
            fieldPath=field_name,
            uniqueCount=field_stats.get("distinct_count"),
            nullCount=field_stats.get("null_count"),
            min=str(field_stats.get("min")) if field_stats.get("min") else None,
            max=str(field_stats.get("max")) if field_stats.get("max") else None,
            mean=str(field_stats.get("mean")) if field_stats.get("mean") else None,
            median=str(field_stats.get("median"))
            if field_stats.get("median")
            else None,
            stdev=str(field_stats.get("stdev")) if field_stats.get("stdev") else None,
            quantiles=[
                QuantileClass(quantile=str(0.25), value=str(quantiles[0])),
                QuantileClass(quantile=str(0.75), value=str(quantiles[1])),
            ]
            if quantiles
            else None,
            sampleValues=field_stats.get("sample_values"),
        )

    def profile_table(self, table_name: str, columns: List[Tuple[str, str]]) -> Dict:
        chunked_columns = self._chunk_columns(columns)
        profile_results = []

        for chunk in chunked_columns:
            try:
                chunk_result = self._profile_chunk(table_name, chunk)
                profile_results.append(chunk_result)
            except DremioAPIException as e:
                self.report.profiling_skipped_other[table_name] += 1
                profile_results.append(self._create_empty_profile_result(chunk))
                self.report.warning(
                    message="Error in profiling for table",
                    context=f"{table_name}",
                    exc=e,
                )
        return self._combine_profile_results(profile_results)

    def _profile_chunk(self, table_name: str, columns: List[Tuple[str, str]]) -> Dict:
        profile_sql = self._build_profile_sql(table_name, columns)
        try:
            results = self.api_operations.execute_query(profile_sql)
            return self._parse_profile_results(results, columns)
        except DremioAPIException as e:
            raise e

    def _chunk_columns(
        self, columns: List[Tuple[str, str]]
    ) -> List[List[Tuple[str, str]]]:
        return [
            columns[i : i + self.MAX_COLUMNS_PER_QUERY]
            for i in range(0, len(columns), self.MAX_COLUMNS_PER_QUERY)
        ]

    def _build_profile_sql(
        self, table_name: str, columns: List[Tuple[str, str]]
    ) -> str:
        metrics = []

        metrics.append("COUNT(*) AS row_count")
        metrics.append(f"{len(columns)} AS column_count")

        if not self.config.profiling.profile_table_level_only:
            for column_name, data_type in columns:
                try:
                    metrics.extend(self._get_column_metrics(column_name, data_type))
                except Exception as e:
                    logger.warning(
                        f"Error building metrics for column {column_name}: {str(e)}"
                    )
                    # Skip this column and continue with others

        if not metrics:
            raise ValueError("No valid metrics could be generated")

        main_query = f"SELECT {', '.join(metrics)} FROM {table_name}"

        if self.config.profiling.limit:
            main_query += f" LIMIT {self.config.profiling.limit}"
        if self.config.profiling.offset:
            main_query += f" OFFSET {self.config.profiling.offset}"

        return main_query

    def _get_column_metrics(self, column_name: str, data_type: str) -> List[str]:
        metrics = []

        # Wrap column name in quotes to handle special characters
        quoted_column_name = f'"{column_name}"'
        safe_column_name = re.sub(r"\W|^(?=\d)", "_", column_name)

        if self.config.profiling.include_field_distinct_count:
            metrics.append(
                f"COUNT(DISTINCT {quoted_column_name}) AS {safe_column_name}_distinct_count"
            )

        if self.config.profiling.include_field_null_count:
            metrics.append(
                f"SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS {safe_column_name}_null_count"
            )

        if self.config.profiling.include_field_min_value:
            metrics.append(f"MIN({quoted_column_name}) AS {safe_column_name}_min")

        if self.config.profiling.include_field_max_value:
            metrics.append(f"MAX({quoted_column_name}) AS {safe_column_name}_max")

        if data_type.lower() in [
            "int",
            "integer",
            "bigint",
            "float",
            "double",
            "decimal",
        ]:
            if self.config.profiling.include_field_mean_value:
                metrics.append(
                    f"AVG(CAST({quoted_column_name} AS DOUBLE)) AS {safe_column_name}_mean"
                )

            if self.config.profiling.include_field_stddev_value:
                metrics.append(
                    f"STDDEV(CAST({quoted_column_name} AS DOUBLE)) AS {safe_column_name}_stdev"
                )

            if self.config.profiling.include_field_median_value:
                metrics.append(
                    f"MEDIAN({quoted_column_name}) AS {safe_column_name}_median"
                )

            if self.config.profiling.include_field_quantiles:
                metrics.append(
                    f"PERCENTILE_CONT(0.25) WITHIN GROUP (ORDER BY {quoted_column_name}) AS {safe_column_name}_25th_percentile"
                )
                metrics.append(
                    f"PERCENTILE_CONT(0.75) WITHIN GROUP (ORDER BY {quoted_column_name}) AS {safe_column_name}_75th_percentile"
                )

        return metrics

    def _parse_profile_results(
        self, results: List[Dict], columns: List[Tuple[str, str]]
    ) -> Dict:
        profile: Dict[str, Any] = {"column_stats": {}}
        result = results[0] if results else {}  # We expect only one row of results

        profile["row_count"] = int(result.get("row_count", 0))

        profile["column_count"] = int(result.get("column_count", 0))

        for column_name, data_type in columns:
            safe_column_name = re.sub(r"\W|^(?=\d)", "_", column_name)
            column_stats: Dict[str, Any] = {}
            if self.config.profiling.include_field_distinct_count:
                null_distinct = result.get(f"{safe_column_name}_distinct_count", 0)
                null_distinct = int(null_distinct) if null_distinct is not None else 0
                column_stats["distinct_count"] = null_distinct

            if self.config.profiling.include_field_null_count:
                null_count_value = result.get(f"{safe_column_name}_null_count", 0)
                null_count = (
                    int(null_count_value) if null_count_value is not None else 0
                )
                column_stats["null_count"] = null_count

            if self.config.profiling.include_field_min_value:
                column_stats["min"] = result.get(f"{safe_column_name}_min")
            if self.config.profiling.include_field_max_value:
                column_stats["max"] = result.get(f"{safe_column_name}_max")

            if data_type.lower() in [
                "int",
                "integer",
                "bigint",
                "float",
                "double",
                "decimal",
            ]:
                if self.config.profiling.include_field_mean_value:
                    column_stats["mean"] = result.get(f"{safe_column_name}_mean")
                if self.config.profiling.include_field_stddev_value:
                    column_stats["stdev"] = result.get(f"{safe_column_name}_stdev")
                if self.config.profiling.include_field_median_value:
                    column_stats["median"] = result.get(f"{safe_column_name}_median")
                if self.config.profiling.include_field_quantiles:
                    column_stats["quantiles"] = [
                        result.get(f"{safe_column_name}_25th_percentile"),
                        result.get(f"{safe_column_name}_75th_percentile"),
                    ]

            profile["column_stats"][column_name] = column_stats

        return profile

    def _create_empty_profile_result(self, columns: List[Tuple[str, str]]) -> Dict:
        profile: Dict[str, Any] = {"column_stats": {}}
        for column_name, _ in columns:
            profile["column_stats"][column_name] = {}
        return profile

    def _combine_profile_results(self, profile_results: List[Dict]) -> Dict:
        combined_profile = {}
        combined_profile["row_count"] = sum(
            profile.get("row_count", 0) for profile in profile_results
        )
        combined_profile["column_count"] = sum(
            profile.get("column_count", 0) for profile in profile_results
        )
        combined_profile["column_stats"] = {}

        for profile in profile_results:
            combined_profile["column_stats"].update(profile.get("column_stats", {}))

        return combined_profile



import logging
import time
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Type, Union

import pandas as pd
from neo4j import GraphDatabase
from pydantic.fields import Field

from datahub.configuration.source_common import EnvConfigMixin
from datahub.emitter.mce_builder import make_data_platform_urn, make_dataset_urn
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import DatasetSubTypes
from datahub.metadata.com.linkedin.pegasus2avro.schema import SchemaFieldDataType
from datahub.metadata.schema_classes import (
    AuditStampClass,
    BooleanTypeClass,
    DatasetPropertiesClass,
    DateTypeClass,
    NullTypeClass,
    NumberTypeClass,
    OtherSchemaClass,
    SchemaFieldClass,
    SchemaMetadataClass,
    StringTypeClass,
    SubTypesClass,
    UnionTypeClass,
)

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

_type_mapping: Dict[Union[Type, str], Type] = {
    "list": UnionTypeClass,
    "boolean": BooleanTypeClass,
    "integer": NumberTypeClass,
    "local_date_time": DateTypeClass,
    "float": NumberTypeClass,
    "string": StringTypeClass,
    "date": DateTypeClass,
    "node": StringTypeClass,
    "relationship": StringTypeClass,
}


class Neo4jConfig(EnvConfigMixin):
    username: str = Field(description="Neo4j Username")
    password: str = Field(description="Neo4j Password")
    uri: str = Field(description="The URI for the Neo4j server")
    env: str = Field(description="Neo4j env")


@dataclass
class Neo4jSourceReport(SourceReport):
    obj_failures: int = 0
    obj_created: int = 0


@platform_name("Neo4j", id="neo4j")
@config_class(Neo4jConfig)
@support_status(SupportStatus.CERTIFIED)
class Neo4jSource(Source):
    NODE = "node"
    RELATIONSHIP = "relationship"
    PLATFORM = "neo4j"

    def __init__(self, ctx: PipelineContext, config: Neo4jConfig):
        self.ctx = ctx
        self.config = config
        self.report = Neo4jSourceReport()

    @classmethod
    def create(cls, config_dict, ctx):
        config = Neo4jConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_field_type(self, attribute_type: Union[type, str]) -> SchemaFieldDataType:
        type_class: type = _type_mapping.get(attribute_type, NullTypeClass)
        return SchemaFieldDataType(type=type_class())

    def get_schema_field_class(
        self, col_name: str, col_type: str, **kwargs: Any
    ) -> SchemaFieldClass:
        if kwargs["obj_type"] == self.NODE and col_type == self.RELATIONSHIP:
            col_type = self.NODE
        else:
            col_type = col_type
        return SchemaFieldClass(
            fieldPath=col_name,
            type=self.get_field_type(col_type),
            nativeDataType=col_type,
            description=col_type.upper()
            if col_type in (self.NODE, self.RELATIONSHIP)
            else col_type,
            lastModified=AuditStampClass(
                time=round(time.time() * 1000), actor="urn:li:corpuser:ingestion"
            ),
        )

    def add_properties(
        self,
        dataset: str,
        description: Optional[str] = None,
        custom_properties: Optional[Dict[str, str]] = None,
    ) -> MetadataChangeProposalWrapper:
        dataset_properties = DatasetPropertiesClass(
            description=description,
            customProperties=custom_properties,
        )
        return MetadataChangeProposalWrapper(
            entityUrn=make_dataset_urn(
                platform=self.PLATFORM, name=dataset, env=self.config.env
            ),
            aspect=dataset_properties,
        )

    def generate_neo4j_object(
        self, dataset: str, columns: list, obj_type: Optional[str] = None
    ) -> MetadataChangeProposalWrapper:
        try:
            fields = [
                self.get_schema_field_class(key, value.lower(), obj_type=obj_type)
                for d in columns
                for key, value in d.items()
            ]
            mcp = MetadataChangeProposalWrapper(
                entityUrn=make_dataset_urn(
                    platform=self.PLATFORM, name=dataset, env=self.config.env
                ),
                aspect=SchemaMetadataClass(
                    schemaName=dataset,
                    platform=make_data_platform_urn(self.PLATFORM),
                    version=0,
                    hash="",
                    platformSchema=OtherSchemaClass(rawSchema=""),
                    lastModified=AuditStampClass(
                        time=round(time.time() * 1000),
                        actor="urn:li:corpuser:ingestion",
                    ),
                    fields=fields,
                ),
            )
            self.report.obj_created += 1
        except Exception as e:
            log.error(e)
            self.report.obj_failures += 1
        return mcp

    def get_neo4j_metadata(self, query: str) -> pd.DataFrame:
        driver = GraphDatabase.driver(
            self.config.uri, auth=(self.config.username, self.config.password)
        )
        """
        This process retrieves the metadata for Neo4j objects using an APOC query, which returns a dictionary
        with two columns: key and value. The key represents the Neo4j object, while the value contains the
        corresponding metadata.

        When data is returned from Neo4j, much of the relationship metadata is stored with the relevant node's
        metadata. Consequently, the objects are organized into two separate dataframes: one for nodes and one for
        relationships.

        In the node dataframe, several fields are extracted and added as new columns. Similarly, in the relationship
        dataframe, certain fields are parsed out, while others require metadata from the nodes dataframe.

        Once the data is parsed and these two dataframes are created, we combine a subset of their columns into a
        single dataframe, which will be used to create the DataHub objects.

        See the docs for examples of metadata:  metadata-ingestion/docs/sources/neo4j/neo4j.md
        """
        try:
            log.info(f"{query}")
            with driver.session() as session:
                result = session.run(query)
                data = [record for record in result]
            log.info("Closing Neo4j driver")
            driver.close()

            node_df = self.process_nodes(data)
            rel_df = self.process_relationships(data, node_df)

            union_cols = ["key", "obj_type", "property_data_types", "description"]
            df = pd.concat([node_df[union_cols], rel_df[union_cols]])
        except Exception as e:
            self.report.failure(
                message="Failed to get neo4j metadata",
                exc=e,
            )

        return df

    def process_nodes(self, data: list) -> pd.DataFrame:
        nodes = [record for record in data if record["value"]["type"] == self.NODE]
        node_df = pd.DataFrame(
            nodes,
            columns=["key", "value"],
        )
        node_df["obj_type"] = node_df["value"].apply(
            lambda record: self.get_obj_type(record)
        )
        node_df["relationships"] = node_df["value"].apply(
            lambda record: self.get_relationships(record)
        )
        node_df["properties"] = node_df["value"].apply(
            lambda record: self.get_properties(record)
        )
        node_df["property_data_types"] = node_df["properties"].apply(
            lambda record: self.get_property_data_types(record)
        )
        node_df["description"] = node_df.apply(
            lambda record: self.get_node_description(record, node_df), axis=1
        )
        return node_df

    def process_relationships(self, data: list, node_df: pd.DataFrame) -> pd.DataFrame:
        rels = [
            record for record in data if record["value"]["type"] == self.RELATIONSHIP
        ]
        rel_df = pd.DataFrame(rels, columns=["key", "value"])
        rel_df["obj_type"] = rel_df["value"].apply(
            lambda record: self.get_obj_type(record)
        )
        rel_df["properties"] = rel_df["value"].apply(
            lambda record: self.get_properties(record)
        )
        rel_df["property_data_types"] = rel_df["properties"].apply(
            lambda record: self.get_property_data_types(record)
        )
        rel_df["description"] = rel_df.apply(
            lambda record: self.get_rel_descriptions(record, node_df), axis=1
        )
        return rel_df

    def get_obj_type(self, record: dict) -> str:
        return record["type"]

    def get_rel_descriptions(self, record: dict, df: pd.DataFrame) -> str:
        descriptions = []
        for _, row in df.iterrows():
            relationships = row.get("relationships", {})
            for relationship, props in relationships.items():
                if record["key"] == relationship:
                    if props["direction"] == "in":
                        for prop in props["labels"]:
                            descriptions.append(
                                f"({row['key']})-[{record['key']}]->({prop})"
                            )
        return "\n".join(descriptions)

    def get_node_description(self, record: dict, df: pd.DataFrame) -> str:
        descriptions = []
        for _, row in df.iterrows():
            if record["key"] == row["key"]:
                for relationship, props in row["relationships"].items():
                    direction = props["direction"]
                    for node in set(props["labels"]):
                        if direction == "in":
                            descriptions.append(
                                f"({row['key']})<-[{relationship}]-({node})"
                            )
                        elif direction == "out":
                            descriptions.append(
                                f"({row['key']})-[{relationship}]->({node})"
                            )

        return "\n".join(descriptions)

    def get_property_data_types(self, record: dict) -> List[dict]:
        return [{k: v["type"]} for k, v in record.items()]

    def get_properties(self, record: dict) -> str:
        return record["properties"]

    def get_relationships(self, record: dict) -> dict:
        return record.get("relationships", None)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        df = self.get_neo4j_metadata(
            "CALL apoc.meta.schema() YIELD value UNWIND keys(value) AS key RETURN key, value[key] AS value;"
        )
        for _, row in df.iterrows():
            try:
                yield MetadataWorkUnit(
                    id=row["key"],
                    mcp=self.generate_neo4j_object(
                        columns=row["property_data_types"],
                        dataset=row["key"],
                    ),
                    is_primary_source=True,
                )

                yield MetadataWorkUnit(
                    id=row["key"],
                    mcp=MetadataChangeProposalWrapper(
                        entityUrn=make_dataset_urn(
                            platform=self.PLATFORM,
                            name=row["key"],
                            env=self.config.env,
                        ),
                        aspect=SubTypesClass(
                            typeNames=[
                                DatasetSubTypes.NEO4J_NODE
                                if row["obj_type"] == self.NODE
                                else DatasetSubTypes.NEO4J_RELATIONSHIP
                            ]
                        ),
                    ),
                )

                yield MetadataWorkUnit(
                    id=row["key"],
                    mcp=self.add_properties(
                        dataset=row["key"],
                        custom_properties=None,
                        description=row["description"],
                    ),
                )

            except Exception as e:
                raise e

    def get_report(self):
        return self.report



from __future__ import annotations

import contextlib
from typing import Iterator

from datahub.utilities.str_enum import StrEnum


class KnownAttribution(StrEnum):
    INGESTION = "INGESTION"
    INGESTION_ALTERNATE = "INGESTION_ALTERNATE"

    UI = "UI"
    SDK = "SDK"

    PROPAGATION = "PROPAGATION"

    def is_ingestion(self) -> bool:
        return self in (
            KnownAttribution.INGESTION,
            KnownAttribution.INGESTION_ALTERNATE,
        )


_default_attribution = KnownAttribution.SDK


def get_default_attribution() -> KnownAttribution:
    return _default_attribution


def set_default_attribution(attribution: KnownAttribution) -> None:
    global _default_attribution
    _default_attribution = attribution


@contextlib.contextmanager
def change_default_attribution(attribution: KnownAttribution) -> Iterator[None]:
    old_attribution = get_default_attribution()
    try:
        set_default_attribution(attribution)
        yield
    finally:
        set_default_attribution(old_attribution)


def is_ingestion_attribution() -> bool:
    return get_default_attribution().is_ingestion()



from typing import IO, Dict, List, Type

from tableschema import Table

from datahub.ingestion.source.schema_inference.base import SchemaInferenceBase
from datahub.metadata.schema_classes import (
    ArrayTypeClass,
    BooleanTypeClass,
    DateTypeClass,
    NullTypeClass,
    NumberTypeClass,
    RecordTypeClass,
    SchemaFieldClass as SchemaField,
    SchemaFieldDataTypeClass as SchemaFieldDataType,
    StringTypeClass,
    TimeTypeClass,
    UnionTypeClass,
)

# see https://github.com/frictionlessdata/tableschema-py/blob/main/tableschema/schema.py#L545
tableschema_type_map: Dict[str, Type] = {
    "duration": TimeTypeClass,
    "geojson": RecordTypeClass,
    "geopoint": RecordTypeClass,
    "object": RecordTypeClass,
    "array": ArrayTypeClass,
    "datetime": TimeTypeClass,
    "time": TimeTypeClass,
    "date": DateTypeClass,
    "integer": NumberTypeClass,
    "number": NumberTypeClass,
    "boolean": BooleanTypeClass,
    "string": StringTypeClass,
    "any": UnionTypeClass,
}


def get_table_schema_fields(table: Table, max_rows: int) -> List[SchemaField]:
    table.infer(limit=max_rows)

    fields: List[SchemaField] = []

    for raw_field in table.schema.fields:
        mapped_type: Type = tableschema_type_map.get(raw_field.type, NullTypeClass)

        field = SchemaField(
            fieldPath=raw_field.name,
            type=SchemaFieldDataType(mapped_type()),
            nativeDataType=str(raw_field.type),
            recursive=False,
        )
        fields.append(field)

    return fields


class CsvInferrer(SchemaInferenceBase):
    def __init__(self, max_rows: int):
        self.max_rows = max_rows

    def infer_schema(self, file: IO[bytes]) -> List[SchemaField]:
        # infer schema of a csv file without reading the whole file
        table = Table(file, format="csv")
        return get_table_schema_fields(table, max_rows=self.max_rows)


class TsvInferrer(SchemaInferenceBase):
    def __init__(self, max_rows: int):
        self.max_rows = max_rows

    def infer_schema(self, file: IO[bytes]) -> List[SchemaField]:
        # infer schema of a tsv file without reading the whole file
        table = Table(file, format="tsv")
        return get_table_schema_fields(table, max_rows=self.max_rows)



import concurrent.futures
import logging
import multiprocessing
from dataclasses import dataclass, field
from functools import partial
from math import ceil
from typing import Callable, Dict, Iterable, List, Optional, Union

from datahub_classify.helper_classes import ColumnInfo, Metadata
from pydantic import Field

from datahub.configuration.common import ConfigModel, ConfigurationError
from datahub.emitter.mce_builder import get_sys_time, make_term_urn, make_user_urn
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.glossary.classifier import ClassificationConfig, Classifier
from datahub.ingestion.glossary.classifier_registry import classifier_registry
from datahub.ingestion.source.common.data_reader import DataReader
from datahub.metadata.com.linkedin.pegasus2avro.common import (
    AuditStamp,
    GlossaryTermAssociation,
    GlossaryTerms,
)
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.com.linkedin.pegasus2avro.schema import SchemaMetadata
from datahub.utilities.lossy_collections import LossyDict, LossyList
from datahub.utilities.perf_timer import PerfTimer

SAMPLE_SIZE_MULTIPLIER = 1.2


logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class ClassificationReportMixin:
    num_tables_fetch_sample_values_failed: int = 0

    num_tables_classification_attempted: int = 0
    num_tables_classification_failed: int = 0
    num_tables_classification_found: int = 0

    info_types_detected: LossyDict[str, LossyList[str]] = field(
        default_factory=LossyDict
    )


class ClassificationSourceConfigMixin(ConfigModel):
    classification: ClassificationConfig = Field(
        default=ClassificationConfig(),
        description="For details, refer to [Classification](../../../../metadata-ingestion/docs/dev_guides/classification.md).",
    )


class ClassificationHandler:
    def __init__(
        self, config: ClassificationSourceConfigMixin, report: ClassificationReportMixin
    ):
        self.config = config
        self.report = report
        self.classifiers = self.get_classifiers()

    def is_classification_enabled(self) -> bool:
        return (
            self.config.classification is not None
            and self.config.classification.enabled
            and len(self.config.classification.classifiers) > 0
        )

    def is_classification_enabled_for_table(self, dataset_name: str) -> bool:
        return (
            self.config.classification is not None
            and self.config.classification.enabled
            and len(self.config.classification.classifiers) > 0
            and self.config.classification.table_pattern.allowed(dataset_name)
        )

    def is_classification_enabled_for_column(
        self, dataset_name: str, column_name: str
    ) -> bool:
        return (
            self.config.classification is not None
            and self.config.classification.enabled
            and len(self.config.classification.classifiers) > 0
            and self.config.classification.table_pattern.allowed(dataset_name)
            and self.config.classification.column_pattern.allowed(
                f"{dataset_name}.{column_name}"
            )
        )

    def get_classifiers(self) -> List[Classifier]:
        classifiers = []

        for classifier in self.config.classification.classifiers:
            classifier_class = classifier_registry.get(classifier.type)
            if classifier_class is None:
                raise ConfigurationError(
                    f"Cannot find classifier class of type={self.config.classification.classifiers[0].type} "
                    " in the registry! Please check the type of the classifier in your config."
                )
            classifiers.append(
                classifier_class.create(
                    config_dict=classifier.config,  # type: ignore
                )
            )

        return classifiers

    def classify_schema_fields(
        self,
        dataset_name: str,
        schema_metadata: SchemaMetadata,
        sample_data: Union[Dict[str, list], Callable[[], Dict[str, list]]],
    ) -> None:
        if not isinstance(sample_data, Dict):
            try:
                # TODO: In future, sample_data fetcher can be lazily called if classification
                # requires values as prediction factor
                sample_data = sample_data()
            except Exception as e:
                self.report.num_tables_fetch_sample_values_failed += 1
                logger.warning(
                    f"Failed to get sample values for dataset. Make sure you have granted SELECT permissions on dataset. {dataset_name}",
                )
                sample_data = dict()
                logger.debug("Error", exc_info=e)

        column_infos = self.get_columns_to_classify(
            dataset_name, schema_metadata, sample_data
        )

        if not column_infos:
            logger.debug(f"No columns in {dataset_name} considered for classification")
            return None

        logger.debug(f"Classifying Table {dataset_name}")

        self.report.num_tables_classification_attempted += 1
        field_terms: Dict[str, str] = {}
        with PerfTimer() as timer:
            try:
                for classifier in self.classifiers:
                    column_infos_with_proposals: Iterable[ColumnInfo]
                    if self.config.classification.max_workers > 1:
                        column_infos_with_proposals = self.async_classify(
                            classifier, column_infos
                        )
                    else:
                        column_infos_with_proposals = classifier.classify(column_infos)

                    for column_info_proposal in column_infos_with_proposals:
                        self.update_field_terms(field_terms, column_info_proposal)

            except Exception:
                self.report.num_tables_classification_failed += 1
                raise
            finally:
                time_taken = timer.elapsed_seconds()
                logger.debug(
                    f"Finished classification {dataset_name}; took {time_taken:.3f} seconds"
                )

        if field_terms:
            self.report.num_tables_classification_found += 1
            self.populate_terms_in_schema_metadata(schema_metadata, field_terms)

    def update_field_terms(
        self, field_terms: Dict[str, str], col_info: ColumnInfo
    ) -> None:
        term = self.get_terms_for_column(col_info)
        if term:
            field_terms[col_info.metadata.name] = term

    def async_classify(
        self, classifier: Classifier, columns: List[ColumnInfo]
    ) -> Iterable[ColumnInfo]:
        num_columns = len(columns)
        BATCH_SIZE = 5  # Number of columns passed to classify api at a time

        logger.debug(
            f"Will Classify {num_columns} column(s) with {self.config.classification.max_workers} worker(s) with batch size {BATCH_SIZE}."
        )

        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.config.classification.max_workers,
            # The fork start method, which is the default on Linux for Python < 3.14, is not
            # safe when the main process uses threads. The default start method on windows/macOS is
            # already spawn, and will be changed to spawn for Linux in Python 3.14.
            # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
            mp_context=multiprocessing.get_context("spawn"),
        ) as executor:
            column_info_proposal_futures = [
                executor.submit(
                    classifier.classify,
                    columns[
                        (i * BATCH_SIZE) : min(i * BATCH_SIZE + BATCH_SIZE, num_columns)
                    ],
                )
                for i in range(ceil(num_columns / BATCH_SIZE))
            ]

            return [
                column_with_proposal
                for proposal_future in concurrent.futures.as_completed(
                    column_info_proposal_futures
                )
                for column_with_proposal in proposal_future.result()
            ]

    def populate_terms_in_schema_metadata(
        self,
        schema_metadata: SchemaMetadata,
        field_terms: Dict[str, str],
    ) -> None:
        for schema_field in schema_metadata.fields:
            if schema_field.fieldPath in field_terms:
                schema_field.glossaryTerms = GlossaryTerms(
                    terms=[
                        GlossaryTermAssociation(
                            urn=make_term_urn(field_terms[schema_field.fieldPath])
                        )
                    ]
                    # Keep existing terms if present
                    + (
                        schema_field.glossaryTerms.terms
                        if schema_field.glossaryTerms
                        else []
                    ),
                    auditStamp=AuditStamp(
                        time=get_sys_time(), actor=make_user_urn("datahub")
                    ),
                )

    def get_terms_for_column(self, col_info: ColumnInfo) -> Optional[str]:
        if not col_info.infotype_proposals:
            return None
        infotype_proposal = max(
            col_info.infotype_proposals, key=lambda p: p.confidence_level
        )
        self.report.info_types_detected.setdefault(
            infotype_proposal.infotype, LossyList()
        ).append(f"{col_info.metadata.dataset_name}.{col_info.metadata.name}")
        term = self.config.classification.info_type_to_term.get(
            infotype_proposal.infotype, infotype_proposal.infotype
        )

        return term

    def get_columns_to_classify(
        self,
        dataset_name: str,
        schema_metadata: SchemaMetadata,
        sample_data: Dict[str, list],
    ) -> List[ColumnInfo]:
        column_infos: List[ColumnInfo] = []

        for schema_field in schema_metadata.fields:
            if not self.is_classification_enabled_for_column(
                dataset_name, schema_field.fieldPath
            ):
                logger.debug(
                    f"Skipping column {dataset_name}.{schema_field.fieldPath} from classification"
                )
                continue

            # As a result of custom field path specification e.g. [version=2.0].[type=struct].[type=struct].service'
            # Sample values for a nested field (an array , union or struct) are not read / passed in classifier correctly.
            # TODO: Fix this behavior for nested fields. This would probably involve:
            # 1. Preprocessing field path spec v2 back to native field representation. (without [*] constructs)
            # 2. Preprocessing retrieved structured sample data to pass in sample values correctly for nested fields.

            column_infos.append(
                ColumnInfo(
                    metadata=Metadata(
                        {
                            "Name": schema_field.fieldPath,
                            "Description": schema_field.description,
                            "DataType": schema_field.nativeDataType,
                            "Dataset_Name": dataset_name,
                        }
                    ),
                    values=(
                        sample_data[schema_field.fieldPath]
                        if schema_field.fieldPath in sample_data.keys()
                        else []
                    ),
                )
            )

        return column_infos


def classification_workunit_processor(
    table_wu_generator: Iterable[MetadataWorkUnit],
    classification_handler: ClassificationHandler,
    data_reader: Optional[DataReader],
    table_id: List[str],
    data_reader_kwargs: Optional[dict] = None,
) -> Iterable[MetadataWorkUnit]:
    """
    Classification handling for a particular table.
    Currently works only for workunits having MCP or MCPW
    """
    table_name = ".".join(table_id)
    if not classification_handler.is_classification_enabled_for_table(table_name):
        yield from table_wu_generator
        return

    for wu in table_wu_generator:
        maybe_schema_metadata = wu.get_aspect_of_type(SchemaMetadata)
        if (
            isinstance(wu.metadata, MetadataChangeEvent)
            and len(wu.metadata.proposedSnapshot.aspects) > 1
        ) or not maybe_schema_metadata:
            yield wu
            continue
        else:  # This is MCP or MCPW workunit with SchemaMetadata aspect
            try:
                classification_handler.classify_schema_fields(
                    table_name,
                    maybe_schema_metadata,
                    (
                        partial(
                            data_reader.get_sample_data_for_table,
                            table_id,
                            classification_handler.config.classification.sample_size
                            * SAMPLE_SIZE_MULTIPLIER,
                            **(data_reader_kwargs or {}),
                        )
                        if data_reader
                        else dict()
                    ),
                )
                yield MetadataChangeProposalWrapper(
                    aspect=maybe_schema_metadata, entityUrn=wu.get_urn()
                ).as_workunit(
                    is_primary_source=wu.is_primary_source,
                )
            except Exception as e:
                logger.warning(
                    f"Failed to classify table columns for {table_name} due to error -> {e}",
                    exc_info=e,
                )
                yield wu



import urllib.parse
from typing import List

# NOTE: Frontend relies on encoding these three characters. Specifically, we decode and encode schema fields for column level lineage.
# If this changes, make appropriate changes to datahub-web-react/src/app/lineage/utils/columnLineageUtils.ts
# We also rely on encoding these exact three characters when generating schemaField urns in our graphQL layer. Update SchemaFieldUtils if this changes.
# Also see https://datahubproject.io/docs/what/urn/#restrictions
RESERVED_CHARS = {",", "(", ")", "␟"}
RESERVED_CHARS_EXTENDED = RESERVED_CHARS.union({"%"})


class UrnEncoder:
    @staticmethod
    def encode_string_array(arr: List[str]) -> List[str]:
        return [UrnEncoder.encode_string(s) for s in arr]

    @staticmethod
    def encode_string(s: str) -> str:
        if not UrnEncoder.contains_reserved_char(s):
            # Fast path for the common case, where no encoding is needed.
            return s
        return "".join(UrnEncoder.encode_char(c) for c in s)

    @staticmethod
    def encode_char(c: str) -> str:
        assert len(c) == 1, "Invalid input, Expected single character"
        return urllib.parse.quote(c) if c in RESERVED_CHARS else c

    @staticmethod
    def contains_reserved_char(value: str) -> bool:
        return bool(set(value).intersection(RESERVED_CHARS))

    @staticmethod
    def contains_extended_reserved_char(value: str) -> bool:
        return bool(set(value).intersection(RESERVED_CHARS_EXTENDED))



from typing import Callable, List, Optional, cast

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import (
    KeyValuePattern,
    TransformerSemantics,
    TransformerSemanticsConfigModel,
)
from datahub.configuration.import_resolver import pydantic_resolve_key
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.transformer.dataset_transformer import (
    DatasetSchemaMetadataTransformer,
)
from datahub.metadata.schema_classes import (
    GlobalTagsClass,
    SchemaFieldClass,
    SchemaMetadataClass,
    TagAssociationClass,
)


class AddDatasetSchemaTagsConfig(TransformerSemanticsConfigModel):
    get_tags_to_add: Callable[[str], List[TagAssociationClass]]

    _resolve_tag_fn = pydantic_resolve_key("get_tags_to_add")


class AddDatasetSchemaTags(DatasetSchemaMetadataTransformer):
    """Transformer that adds glossary tags to datasets according to a callback function."""

    ctx: PipelineContext
    config: AddDatasetSchemaTagsConfig

    def __init__(self, config: AddDatasetSchemaTagsConfig, ctx: PipelineContext):
        super().__init__()
        self.ctx = ctx
        self.config = config

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "AddDatasetSchemaTags":
        config = AddDatasetSchemaTagsConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def extend_field(
        self, schema_field: SchemaFieldClass, server_field: Optional[SchemaFieldClass]
    ) -> SchemaFieldClass:
        all_tags = self.config.get_tags_to_add(schema_field.fieldPath)
        if len(all_tags) == 0:
            # return input schema field as there is no tags to add
            return schema_field

        # Add existing tags if user want to keep them
        if (
            schema_field.globalTags is not None
            and self.config.replace_existing is False
        ):
            all_tags.extend(schema_field.globalTags.tags)

        tags_to_add: List[TagAssociationClass] = []
        server_tags: List[TagAssociationClass] = []
        if server_field is not None and server_field.globalTags is not None:
            # Go for patch
            server_tags = server_field.globalTags.tags
            server_tag_urns = [tag_association.tag for tag_association in server_tags]
            for tag in all_tags:
                if tag.tag not in server_tag_urns:
                    tags_to_add.append(tag)

        # Set tags_to_add to all_tags if server tags were not present
        if len(tags_to_add) == 0:
            tags_to_add = all_tags

        new_global_tag_class: GlobalTagsClass = GlobalTagsClass(tags=[])
        new_global_tag_class.tags.extend(tags_to_add)
        new_global_tag_class.tags.extend(server_tags)

        schema_field.globalTags = new_global_tag_class

        return schema_field

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[builder.Aspect]
    ) -> Optional[builder.Aspect]:
        schema_metadata_aspect: SchemaMetadataClass = cast(SchemaMetadataClass, aspect)
        assert schema_metadata_aspect is None or isinstance(
            schema_metadata_aspect, SchemaMetadataClass
        )

        server_field_map: dict = {}
        if self.config.semantics == TransformerSemantics.PATCH:
            assert self.ctx.graph
            server_schema_metadata_aspect: Optional[SchemaMetadataClass] = (
                self.ctx.graph.get_schema_metadata(entity_urn=entity_urn)
            )
            if server_schema_metadata_aspect is not None:
                if not schema_metadata_aspect:
                    schema_metadata_aspect = server_schema_metadata_aspect

                input_field_path = [
                    field.fieldPath
                    for field in schema_metadata_aspect.fields
                    if field is not None
                ]
                # cache the server field to use in patching the schema-field later
                for field in server_schema_metadata_aspect.fields:
                    server_field_map[field.fieldPath] = field
                    if field.fieldPath not in input_field_path:
                        # This field is present on server but not in input aspect
                        schema_metadata_aspect.fields.append(field)

        if not schema_metadata_aspect:
            # We can't add tags to a schema that doesn't exist.
            return None

        schema_metadata_aspect.fields = [
            self.extend_field(field, server_field=server_field_map.get(field.fieldPath))
            for field in schema_metadata_aspect.fields
        ]

        return schema_metadata_aspect  # type: ignore


class PatternDatasetTagsConfig(TransformerSemanticsConfigModel):
    tag_pattern: KeyValuePattern = KeyValuePattern.all()


class PatternAddDatasetSchemaTags(AddDatasetSchemaTags):
    """Transformer that adds a dynamic set of tags to each field in a dataset based on supplied patterns."""

    def __init__(self, config: PatternDatasetTagsConfig, ctx: PipelineContext):
        tag_pattern = config.tag_pattern
        generic_config = AddDatasetSchemaTagsConfig(
            get_tags_to_add=lambda path: [
                TagAssociationClass(tag=urn) for urn in tag_pattern.value(path)
            ],
            semantics=config.semantics,
            replace_existing=config.replace_existing,
        )
        super().__init__(generic_config, ctx)

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "PatternAddDatasetSchemaTags":
        config = PatternDatasetTagsConfig.parse_obj(config_dict)
        return cls(config, ctx)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, overload

from datahub.errors import ItemNotFoundError, MultipleItemsFoundError, SdkUsageError
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.urns import (
    CorpUserUrn,
    DomainUrn,
    GlossaryTermUrn,
)

if TYPE_CHECKING:
    from datahub.sdk.main_client import DataHubClient


class ResolverClient:
    def __init__(self, client: DataHubClient):
        self._client = client

    # TODO: add caching to this method

    @property
    def _graph(self) -> DataHubGraph:
        return self._client._graph

    def domain(self, *, name: str) -> DomainUrn:
        urn_str = self._graph.get_domain_urn_by_name(name)
        if urn_str is None:
            raise ItemNotFoundError(f"Domain with name {name} not found")
        return DomainUrn.from_string(urn_str)

    @overload
    def user(self, *, name: str) -> CorpUserUrn: ...
    @overload
    def user(self, *, email: str) -> CorpUserUrn: ...
    def user(
        self, *, name: Optional[str] = None, email: Optional[str] = None
    ) -> CorpUserUrn:
        filter_explanation: str
        filters = []
        if name is not None:
            if email is not None:
                raise SdkUsageError("Cannot specify both name and email for auto_user")
            # TODO: do we filter on displayName or fullName?
            filter_explanation = f"with name {name}"
            filters.append(
                {
                    "field": "fullName",
                    "values": [name],
                    "condition": "EQUAL",
                }
            )
        elif email is not None:
            filter_explanation = f"with email {email}"
            filters.append(
                {
                    "field": "email",
                    "values": [email],
                    "condition": "EQUAL",
                }
            )
        else:
            raise SdkUsageError("Must specify either name or email for auto_user")

        users = list(
            self._graph.get_urns_by_filter(
                entity_types=[CorpUserUrn.ENTITY_TYPE],
                extraFilters=filters,
            )
        )
        if len(users) == 0:
            # TODO: In auto methods, should we just create the user/domain/etc if it doesn't exist?
            raise ItemNotFoundError(f"User {filter_explanation} not found")
        elif len(users) > 1:
            raise MultipleItemsFoundError(
                f"Multiple users found {filter_explanation}: {users}"
            )
        else:
            return CorpUserUrn.from_string(users[0])

    def term(self, *, name: str) -> GlossaryTermUrn:
        # TODO: Add some limits on the graph fetch
        terms = list(
            self._graph.get_urns_by_filter(
                entity_types=[GlossaryTermUrn.ENTITY_TYPE],
                extraFilters=[
                    {
                        "field": "id",
                        "values": [name],
                        "condition": "EQUAL",
                    }
                ],
            )
        )
        if len(terms) == 0:
            raise ItemNotFoundError(f"Term with name {name} not found")
        elif len(terms) > 1:
            raise SdkUsageError(f"Multiple terms found with name {name}: {terms}")
        else:
            return GlossaryTermUrn.from_string(terms[0])



import dataclasses
from dataclasses import field as dataclass_field

from datahub.ingestion.api.source import SourceReport
from datahub.utilities.lossy_collections import LossyList


@dataclasses.dataclass
class DeltaLakeSourceReport(SourceReport):
    files_scanned = 0
    filtered: LossyList[str] = dataclass_field(default_factory=LossyList)

    def report_file_scanned(self) -> None:
        self.files_scanned += 1

    def report_file_dropped(self, file: str) -> None:
        self.filtered.append(file)
